/*************************************************************************
 *                        Boar 's Only A Raytracer                       *
 *                                                                       *
 *       Copyright (C) 2014  Holger Bettag  (hobold@vectorizer.org)      *
 *                                                                       *
 * This program is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation, either version 3 of the License, or     *
 * (at your option) any later version.                                   *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *************************************************************************/

#include <stdio.h>
#include <stdint.h>
#include <sys/time.h>

#include "boar.h"

const int ImgWidth = 800;
const int ImgHeight = 500;

const int Frames = 400;

const int tileSize = 11;
const int tileSizeAA = 14;

#if defined(STATISTICS)
// global statistics counters
int64_t numRays = 0;         // number of view rays
int64_t numShadow = 0;       // number of shadow rays
// more global statistics
int64_t numBranch = 0;       // number of visited KdTree branches
int64_t numLeaf = 0;         // number of visited KdTree leaves
int64_t numQuadric = 0;      // number of tested quadric surfaces
int64_t numIntersection = 0; // number of tested CSG intersections
int64_t numIntersectionStep = 0;  // number of iterations in CSG intersection
int64_t numQIntersection = 0;// number of tested QuadricIntersections
int64_t numUnion = 0;        // number of tested CSG unions
int64_t numStitch = 0;       // number of tested CSG stitches
#endif  // defined(STATISTICS)

// time keeping
double begintime, endtime;

// render one tile, called by renderFrameTiled()
void renderTile(Scene* scene, Camera* cam, int xMin, int yMin,
		float* buffer) {
  Ray ray;
  Vec3 colour;
#if defined(ANTIALIAS)
  uintptr_t hash;
#endif  // defined(ANTIALIAS)
  
  int xMax = xMin + tileSize;  // determine clipped tile size
  if (xMax >= ImgWidth) {
    xMax = ImgWidth;
  }
  int yMax = yMin + tileSize;
  if (yMax >= ImgHeight) {
    yMax = ImgHeight;
  }
  
  for (int y = yMin; y < yMax; y++) {
    double fy = (double(ImgHeight - 1 - y) + 0.5)/double(ImgHeight);
    for (int x = xMin; x < xMax; x++) {
      ray = cam->makeRay((double(x) + 0.5)/double(ImgWidth), fy);
#if defined(ANTIALIAS)
      hash = 0;
#endif  // defined(ANTIALIAS)
      colour = TraceRay(ray, *scene
#if defined(ANTIALIAS)
			, hash
#endif  // defined(ANTIALIAS)
			);
    
      buffer[0] = qmin(colour.x, 1.0);  // clamp intensities
      buffer[1] = qmin(colour.y, 1.0);
      buffer[2] = qmin(colour.z, 1.0);
      
#if defined(ANTIALIAS)
      ((uint32_t*)(buffer))[3] = (uint32_t)(hash ^ (hash >> 32));
#endif  // defined(ANTIALIAS)
      buffer += 4;          // skip to next pixel
    }
    buffer += (ImgWidth - 1)*4;  // skip to next row
    y++;
    if (y >= yMax) {         // check for loop end
      break;
    }
    fy = (double(ImgHeight - 1 - y) + 0.5)/double(ImgHeight);
    for (int x = xMax - 1; x >= xMin; x--) {
      ray = cam->makeRay((double(x) + 0.5)/double(ImgWidth), fy);
#if defined(ANTIALIAS)
      hash = 0;
#endif  // defined(ANTIALIAS)
      colour = TraceRay(ray, *scene
#if defined(ANTIALIAS)
			, hash
#endif  // defined(ANTIALIAS)
			);
    
      buffer[0] = qmin(colour.x, 1.0);  // clamp intensities
      buffer[1] = qmin(colour.y, 1.0);
      buffer[2] = qmin(colour.z, 1.0);

#if defined(ANTIALIAS)
      ((uint32_t*)(buffer))[3] = (uint32_t)(hash ^ (hash >> 32));
#endif  // defined(ANTIALIAS)
      buffer -= 4;          // skip to next pixel
    }
    buffer += (ImgWidth + 1)*4;  // skip to next row    
  }
  return;
}


#if defined(ANTIALIAS)

// supersample pixels of one Tile when hash is nonzero
void supersampleTile(Scene* scene, Camera* cam, int xMin, int yMin,
		     float* buffer) {
  Ray ray;
  Vec3 colourSum, colour;
  uintptr_t hash = 0;

  int xMax = xMin + tileSizeAA;  // determine clipped tile size
  if (xMax >= ImgWidth) {
    xMax = ImgWidth;
  }
  int yMax = yMin + tileSizeAA;
  if (yMax >= ImgHeight) {
    yMax = ImgHeight;
  }

  for (int y = yMin; y < yMax; y++) {
    for (int x = xMin; x < xMax; x++) {
      if (((int32_t*)(buffer))[0] < 0) {
		
	// read back colour of center ray
	colourSum.x = -buffer[0];  // fix up flagged sign bit
	colourSum.y = buffer[1];
	colourSum.z = buffer[2];
	
	ray = cam->makeRay((double(x) + 0.7)/double(ImgWidth),
			   (double(ImgHeight-1 - y) + 0.1)/double(ImgHeight));
	colour = TraceRay(ray, *scene, hash);      
	colourSum.x += qmin(colour.x, 1.0);  // add clamped intensities
	colourSum.y += qmin(colour.y, 1.0);
	colourSum.z += qmin(colour.z, 1.0);
	
	ray = cam->makeRay((double(x) + 0.1)/double(ImgWidth),
			   (double(ImgHeight-1 - y) + 0.3)/double(ImgHeight));
	colour = TraceRay(ray, *scene, hash);      
	colourSum.x += qmin(colour.x, 1.0);
	colourSum.y += qmin(colour.y, 1.0);
	colourSum.z += qmin(colour.z, 1.0);
	
	ray = cam->makeRay((double(x) + 0.9)/double(ImgWidth),
			   (double(ImgHeight-1 - y) + 0.7)/double(ImgHeight));
	colour = TraceRay(ray, *scene, hash);      
	colourSum.x += qmin(colour.x, 1.0);
	colourSum.y += qmin(colour.y, 1.0);
	colourSum.z += qmin(colour.z, 1.0);
	
	ray = cam->makeRay((double(x) + 0.3)/double(ImgWidth),
			   (double(ImgHeight-1 - y) + 0.9)/double(ImgHeight));
	colour = TraceRay(ray, *scene, hash);      
	colourSum.x += qmin(colour.x, 1.0);
	colourSum.y += qmin(colour.y, 1.0);
	colourSum.z += qmin(colour.z, 1.0);
	
	buffer[0] = colourSum.x * 0.2;  // divide by 5.0
	buffer[1] = colourSum.y * 0.2;
	buffer[2] = colourSum.z * 0.2;
      }
      buffer += 4;
    }
    buffer += (ImgWidth - (xMax - xMin))*4;
  }
  return;
}

#endif  // defined(ANTIALIAS)


#if defined(ANTIALIAS)

// compute flag from hash bits, store into sign of red colour component
// this routine does NOT work for the topmost and bottommost rows
// (pointer must be adjusted to point to 4th colour component)
void prepareAAflagRow(uint32_t* pixels) {
  uint32_t flag;
  int x;
  
  // left border: five neighbours
  flag = pixels[0] ^ pixels[(-ImgWidth)*4];        // top
  flag |= pixels[0] ^ pixels[(1 - ImgWidth)*4];    // top right
  flag |= pixels[0] ^ pixels[(1)*4];               // right
  flag |= pixels[0] ^ pixels[(ImgWidth)*4];        // bottom
  flag |= pixels[0] ^ pixels[(ImgWidth + 1)*4];    // bottom right
  
  flag = (flag != 0) << 31;  // move bit flag into sign bit
  pixels[-3] |= flag;        // store in sign of first colour component

  // loop over non-border pixels
  for (x = 1; x < ImgWidth - 1; x++) {
    // all eight neighbours
    flag = pixels[x*4] ^ pixels[(x - ImgWidth - 1)*4];   // top left
    flag |= pixels[x*4] ^ pixels[(x - ImgWidth)*4];      // top
    flag |= pixels[x*4] ^ pixels[(x + 1 - ImgWidth)*4];  // top right
    flag |= pixels[x*4] ^ pixels[(x - 1)*4];             // left
    flag |= pixels[x*4] ^ pixels[(x + 1)*4];             // right
    flag |= pixels[x*4] ^ pixels[(x + ImgWidth - 1)*4];  // bottom left
    flag |= pixels[x*4] ^ pixels[(x + ImgWidth)*4];      // bottom
    flag |= pixels[x*4] ^ pixels[(x + ImgWidth + 1)*4];  // bottom right
    
    flag = (flag != 0) << 31;  // move boolean flag into sign bit
    pixels[x*4 - 3] |= flag;   // store in sign of first colour component
  }
  
  // right border: five neighbours
  flag = pixels[x*4] ^ pixels[(x - ImgWidth - 1)*4];   // top left
  flag |= pixels[x*4] ^ pixels[(x - ImgWidth)*4];      // top
  flag |= pixels[x*4] ^ pixels[(x - 1)*4];             // left
  flag |= pixels[x*4] ^ pixels[(x + ImgWidth - 1)*4];  // bottom left
  flag |= pixels[x*4] ^ pixels[(x + ImgWidth)*4];      // bottom
  
  flag = (flag != 0) << 31;  // move bit flag into sign bit
  pixels[x*4 - 3] |= flag;   // store in sign of first colour component

  return;
}

#endif  // defined(ANTIALIAS)

/*
#if defined(ANTIALIAS)

// replace hash bits with a boolean flag that is true whenever a pixel's
// hash differs from the hash of any adjacent (8 neighbours) pixel, and
// false otherwise
void prepareAAflags(uint32_t* pixels) {
  int x, y;
  uint32_t* buffer = new uint32_t[ImgWidth*3];

  if (buffer == NULL) {
    fprintf(stderr,"out of memory for convolution buffer\n");
    exit(EXIT_FAILURE);
  }

  for (x = 0; x < ImgWidth; x++) {
    buffer[x] = pixels[x*4 + 3];
    buffer[ImgWidth + x] = pixels[x*4 + 3];
  }
  
  uint32_t flag;
  
  for (y = 0; y < ImgHeight - 1; y++) {
    // fill bottom row of convolution buffer
    for (x = 0; x < ImgWidth; x++) {
      buffer[2*ImgWidth + x] = pixels[((y+1)*ImgWidth + x)*4 + 3];
    }
    
    // left border: six neighbours
    flag = (buffer[ImgWidth] != buffer[0]);               // top
    flag |= (buffer[ImgWidth] != buffer[1]);              // top right
    flag |= (buffer[ImgWidth] != buffer[ImgWidth + 1]);   // right
    flag |= (buffer[ImgWidth] != buffer[2*ImgWidth]);     // bottom
    flag |= (buffer[ImgWidth] != buffer[2*ImgWidth + 1]); // bottom right
    pixels[y*ImgWidth*4 + 3] = flag;

    // loop over non-border pixels
    for (x = 1; x < ImgWidth - 1; x++) {
      flag = (buffer[ImgWidth + x] != buffer[x - 1]);  // top left
      flag |= (buffer[ImgWidth + x] != buffer[x]);  // top
      flag |= (buffer[ImgWidth + x] != buffer[x + 1]);  // top right
      flag |= (buffer[ImgWidth + x] != buffer[ImgWidth + x - 1]);  // left
      flag |= (buffer[ImgWidth + x] != buffer[ImgWidth + x + 1]);  // right
      flag |= (buffer[ImgWidth + x] != buffer[2*ImgWidth + x-1]); // bottom left
      flag |= (buffer[ImgWidth + x] != buffer[2*ImgWidth + x]);  // bottom
      flag |= (buffer[ImgWidth + x] != buffer[2*ImgWidth+x+1]); // bottom right
      pixels[(y*ImgWidth + x)*4 + 3] = flag;
    }

    // right border
    flag = (buffer[ImgWidth + x] != buffer[x - 1]);  // top left
    flag |= (buffer[ImgWidth + x] != buffer[x]);  // top
    flag |= (buffer[ImgWidth + x] != buffer[ImgWidth + x - 1]);  // left
    flag |= (buffer[ImgWidth + x] != buffer[2*ImgWidth + x-1]); // bottom left
    flag |= (buffer[ImgWidth + x] != buffer[2*ImgWidth + x]);  // bottom
    pixels[(y*ImgWidth + x)*4 + 3] = flag;
    
    // scroll convolution buffer upwards
    for (x = 0; x < ImgWidth; x++) {
      buffer[x] = buffer[ImgWidth + x];
      buffer[ImgWidth + x] = buffer[2*ImgWidth + x];
    }
  }

  // bottom row
  // left border: six neighbours
  flag = (buffer[ImgWidth] != buffer[0]);               // top
  flag |= (buffer[ImgWidth] != buffer[1]);              // top right
  flag |= (buffer[ImgWidth] != buffer[ImgWidth + 1]);   // right
  pixels[y*ImgWidth*4 + 3] = flag;
  
  // loop over non-border pixels
  for (x = 1; x < ImgWidth - 1; x++) {
    flag = (buffer[ImgWidth + x] != buffer[x - 1]);  // top left
    flag |= (buffer[ImgWidth + x] != buffer[x]);  // top
    flag |= (buffer[ImgWidth + x] != buffer[x + 1]);  // top right
    flag |= (buffer[ImgWidth + x] != buffer[ImgWidth + x - 1]);  // left
    flag |= (buffer[ImgWidth + x] != buffer[ImgWidth + x + 1]);  // right
    pixels[(y*ImgWidth + x)*4 + 3] = flag;
  }
  
  // right border
  flag = (buffer[ImgWidth + x] != buffer[x - 1]);  // top left
  flag |= (buffer[ImgWidth + x] != buffer[x]);  // top
  flag |= (buffer[ImgWidth + x] != buffer[ImgWidth + x - 1]);  // left
  pixels[(y*ImgWidth + x)*4 + 3] = flag;

  return;
}

#endif  // defined(ANTIALIAS)
*/

// render one frame, called by renderFrames()
void renderFrameTiled(Scene* scene, Camera* cam, char* fileName) {
  float* fpixels = new float[ImgWidth*ImgHeight*4];

  //int x, y;
  
  if (fpixels == NULL) {
    fprintf(stderr,"out of memory for frame buffer\n");
    exit(EXIT_FAILURE);
  }

  // distribute image tiles as parallel tasks


  // wait for all rows to finish

  
#if defined(ANTIALIAS)

  // flag pixels that need supersampling
  //prepareAAflags((uint32_t*)fpixels);

  
 

  // distribute supersampling tiles as parallel tasks


  // wait for all tasks to finish


  // ***** ANTIALIASING PASS ENDS HERE *****
  
#endif  // defined(ANTIALIAS)


  FILE* outImageFile = fopen(fileName, "wb");
  if (outImageFile == NULL) {
    fprintf(stderr,"could not open output file %s\n", fileName);
    exit(EXIT_FAILURE);
  }

  // simple raw PPM header
  fprintf(outImageFile,"P6\n%d %d\n255\n", ImgWidth, ImgHeight);
  // convert image data to RGB 8:8:8
  float* current = fpixels;
  uint8_t* dest = (uint8_t*)fpixels;
  for (int y = ImgHeight; y > 0; y--) {
    for (int x = ImgWidth; x > 0; x--) {
      double frac = 0.25*double((x&1) + (((x^y)&1) << 1));
      
      dest[0] = uint8_t(current[0]*255.0 + frac);
      dest[1] = uint8_t(current[1]*255.0 + frac);
      dest[2] = uint8_t(current[2]*255.0 + frac);
      /*
      putc(int(current[0]*255.999), outImageFile);
      putc(int(current[1]*255.999), outImageFile);
      putc(int(current[2]*255.999), outImageFile);
      */

      current += 4;
      dest += 3;
    }
  }
  // write image data
  fwrite(fpixels, ImgWidth*ImgHeight*3, 1, outImageFile);
  
  fclose(outImageFile);
  
  delete[] fpixels;

  return;
}

// render simple fly around
void renderFrames(Scene* scene) {
  Camera* cam = scene->camera;
  char fileName[256];
  double angle;
  const ListNode* curLight;     // list item: light
  PointLight* theLight;  // actual light object
  Transform rotation;
  rotation.RotateY(-360.0/double(Frames));

  
  Transform rotationB;
  rotationB.RotateY(100.0 * 360.0/double(Frames));
  // rotate lights
  curLight = scene->lights.firstNode();


  while (curLight != NULL) {
    theLight = (PointLight*)(curLight->getData());
    theLight->position = rotationB.ApplyTo(theLight->position);
    curLight = curLight->nextNode();
  }
  

  for (int i = 0; i < Frames; i++) {

    sprintf(fileName, "frames/frame_%05d.ppm", i);
    
    // rotate camera
    angle = double(-i)*360.0/double(Frames);
    //angle = double(Frames - i)*360.0/double(Frames);
    ((PerspectiveCam*)cam)->InitRotatedLookAt(angle);

    
    // rotate lights
    curLight = scene->lights.firstNode();

    while (curLight != NULL) {
      theLight = (PointLight*)(curLight->getData());
      theLight->position = rotation.ApplyTo(theLight->position);
      curLight = curLight->nextNode();

      /*
      if (curLight != NULL) {
	curLight = curLight->nextNode();  // skip every other light
      }
      */

    }
    

    renderFrameTiled(scene, cam, fileName);
    fprintf(stderr,"\rrendered frame %d  ", i);    
  }

  return;
}

// returns gettimeofday converted to a floating point value
double myGetTime(void) {
  timeval now;

  if (gettimeofday(&now, NULL) != 0) {
    fprintf(stderr,"warning: gettimeofday() failed.\n");
  }
  
  return double(now.tv_sec) + double(now.tv_usec)/1e6;
}

// output formatted time string to stderr
void printMyTime(double Time) {
  bool printZeros = false;   // skip leading zeros, print inner zeros
  double howmany = 0.0;      // number of current units

  if (Time >= 3600) {        // an hour or longer?
    // print number of hours
    printZeros = true;
    howmany = floor(Time / 3600.0);
    Time -= howmany * 3600.0;
    fprintf(stderr,"%dh", int(howmany));
  }

  if (printZeros || (Time >= 60)) {  // a minute or longer?
    // print number of minutes
    printZeros = true;
    howmany = floor(Time / 60.0);
    Time -= howmany * 60.0;
    fprintf(stderr,"%dm", int(howmany));
  }
  
  // print number of seconds
  fprintf(stderr,"%5.3fs\n", Time);
  
  return;
}

int main(int argc, char** argv) {
  Scene* scene = NULL;
  
  if (argc != 2) {
    fprintf(stderr,"Error: no scene file name given\n");
    return 1;
  }

  ParserStack parser = ParserStack(argv[1]);
  
  begintime = myGetTime();
  scene = parser.parseScene();
  endtime = myGetTime();
  printMyTime(endtime - begintime);

  /*
  begintime = myGetTime();
  scene->root->buildSpatialSubdivision();

  scene->kDtree->transcodeVoxels(scene->root);
  endtime = myGetTime();
  printMyTime(endtime - begintime);
  */


  begintime = myGetTime();

  /*
  // schedule saving and termination after computation
  dispatch_group_notify(global_group, frame_queue, ^{
      endtime = myGetTime();
      fprintf(stderr, "\rrendering finished                   \n");
      printMyTime(endtime - begintime);
#if defined(STATISTICS)
      unsigned long totalRays = numRays + numShadow;
      fprintf(stderr, "%lu view rays, %lu shadow rays (%lu total)\n",
	      (unsigned long)numRays, (unsigned long)numShadow, totalRays);
      fprintf(stderr, "visited %lu branches and %lu leaves\n"
	      "(%10.8f and %10.8f per ray)\n",
	      (unsigned long)numBranch, (unsigned long)numLeaf,
	      double(numBranch)/double(totalRays),
	      double(numLeaf)/double(totalRays));
      fprintf(stderr, "tested %lu quadrics (%10.8f per leaf),\n",
	      (unsigned long)numQuadric, double(numQuadric)/double(numLeaf));
      fprintf(stderr, "%lu CSG intersections (%10.8f per leaf), %lu steps,\n",
	      (unsigned long)numIntersection,
	      double(numIntersection)/double(numLeaf),
	      (unsigned long)numIntersectionStep);
      fprintf(stderr, "%lu QuadricIntersections (%10.8f per leaf),\n",
	      (unsigned long)numQIntersection,
	      double(numQIntersection)/double(numLeaf));
      fprintf(stderr, "%lu CSG unions (%10.8f per leaf),\n",
	      (unsigned long)numUnion, double(numUnion)/double(numLeaf));
      fprintf(stderr, "and %lu CSG stitches (%10.8f per leaf).\n",
	      (unsigned long)numStitch, double(numStitch)/double(numLeaf));
#endif
      exit(0);
    });
  */

  return 0;
}
