/*************************************************************************
 *                        Boar 's Only A Raytracer                       *
 *                                                                       *
 *       Copyright (C) 2014  Holger Bettag  (hobold@vectorizer.org)      *
 *                                                                       *
 * This program is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation, either version 3 of the License, or     *
 * (at your option) any later version.                                   *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *************************************************************************/

#include <stdio.h>

#include "camera.h"
#include "transform.h"


// camera setup using "look at" parameters
// (implies right handed coordinate system)
// internal use only
void PerspectiveCam::internalLookAt(Vec3 Eye,        // observer position
				    Vec3 LookAt,     // desired center of image
				    Vec3 Sky,        // reference direction
				    double Width,    // view rect width
				    double Height,   // view rect height
				    double Distance) // view rect distance
{  
  // make LookAt relative from Eye
  LookAt.x -= Eye.x;
  LookAt.y -= Eye.y;
  LookAt.z -= Eye.z;
    
  // set length of LookAt to Distance
  LookAt.setLen(Distance);  // now points from eye to center of view rect

  /*
  fprintf(stderr,"Cam: viewDir(%f,%f,%f)\n",
	  LookAt.x, LookAt.y, LookAt.z);
  */  

  // find viewRight
  viewRight.crossOf(LookAt, Sky); // perpendicular to both vectors
  viewRight.setLen(Width);        // points from left to right edge

  /*
  fprintf(stderr,"Cam: viewRight(%f,%f,%f)\n",
	  viewRight.x, viewRight.y, viewRight.z);
  */

  // find viewUp
  viewUp.crossOf(viewRight, LookAt); // again perpendicular to both vectors
  viewUp.setLen(Height);             // points from bottom to top edge

  /*
  fprintf(stderr,"Cam: viewUp(%f,%f,%f)\n",
	  viewUp.x, viewUp.y, viewUp.z);
  */

  // base point of view rect is lower left corner, so adjust from center
  viewBase.x = Eye.x + LookAt.x - 0.5*(viewRight.x + viewUp.x);
  viewBase.y = Eye.y + LookAt.y - 0.5*(viewRight.y + viewUp.y);
  viewBase.z = Eye.z + LookAt.z - 0.5*(viewRight.z + viewUp.z);

  /*
  fprintf(stderr,"Cam: viewBase(%f,%f,%f)\n",
	  viewBase.x, viewBase.y, viewBase.z);
  */

  // remember eye point
  eye = Eye;
  
  return;
}


// camera setup using "look at" parameters
// (implies right handed coordinate system)
void PerspectiveCam::InitLookAt(Vec3 Eye,        // observer position
				Vec3 LookAt,     // desired center of image
				Vec3 Sky,        // reference direction
				double Width,    // view rect width
				double Height,   // view rect height
				double Distance) // view rect distance
{
  // remember LookAt values, for InitRotatedLookAt
  lookAt = LookAt;
  sky = Sky;
  width = Width;
  height = Height;
  distance = Distance;
  originalEye = Eye;

  internalLookAt(Eye, LookAt, Sky, Width, Height, Distance);
  
  return;
}

// for quick and dirty camera flights after InitLookAt()
// rotate around LookAt point, axis parallel to Y axis
// angle in degrees
void PerspectiveCam::InitRotatedLookAt(double angle) {
  Vec3 newEye;
  Transform rotation;
  
  newEye = originalEye;
  newEye.subtract(lookAt);  // move LookAt to coordinate origin
  
  rotation.RotateY(angle);  // construct rotation matrix

  newEye = rotation.ApplyTo(newEye);  // rotate eye position

  newEye.add(lookAt);  // move coordinate origin back to LookAt
  
  // initialize camera for new position with old parameters
  internalLookAt(newEye, lookAt, sky, width, height, distance);

  return;
}


// generate ray corresponding to abstract screen coordinates
// (coordinates range from 0.0 to 1.0 inclusive, origin is lower left)
Ray PerspectiveCam::makeRay(double x, double y) const {
  Ray result;
  Vec3 dir;

  result.base = eye;
  
  dir.x = viewBase.x + x*viewRight.x + y*viewUp.x - eye.x;
  dir.y = viewBase.y + x*viewRight.y + y*viewUp.y - eye.y;
  dir.z = viewBase.z + x*viewRight.z + y*viewUp.z - eye.z;

  result.dir.normalize(dir);
  
  return result;
}
