/*************************************************************************
 *                        Boar 's Only A Raytracer                       *
 *                                                                       *
 *       Copyright (C) 2014  Holger Bettag  (hobold@vectorizer.org)      *
 *                                                                       *
 * This program is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation, either version 3 of the License, or     *
 * (at your option) any later version.                                   *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#if defined(STATISTICS)
#include <libkern/OSAtomic.h>
#endif

#include "csg.h"

#if defined(STATISTICS)
// global statistics counters, defined in boar.c++
extern int64_t numIntersection;
extern int64_t numQIntersection;
extern int64_t numIntersectionStep;
extern int64_t numUnion;
extern int64_t numStitch;
#endif

// something slightly larger than zero, perhaps?
//const double CSG_EPSILON_FUDGE = 0.00005;
const double CSG_EPSILON_FUDGE = 0.0;

Material* CSGObject::getMaterial(void) const {
  // only return material when it is shared by both children
  if (child1->getMaterial() == child2->getMaterial())
    return child1->getMaterial();

  return NULL;  // otherwise return zero
}

void CSGObject::setMaterial(const Material* Mat, bool forceReplace) {
  child1->setMaterial(Mat, forceReplace);
  child2->setMaterial(Mat, forceReplace);
}

int CSGObject::countPrimitives(void) {
  if (numPrimitives == 0) {
    numPrimitives = child1->countPrimitives() + child2->countPrimitives();
  }
  return numPrimitives;
}

void CSGObject::applyTransform(const Transform& trans) {
  child1->applyTransform(trans);
  child2->applyTransform(trans);
}

void CSGObject::insideOut(void) {
  fprintf(stderr,"inside out inversion not implemented for CSG");
  exit(EXIT_FAILURE);
}

Intersection::Intersection(GeObject* Child1, GeObject* Child2) {
  child1 = Child1;
  child2 = Child2;
  numPrimitives = 0;  // clear cache
}

VoxClass Intersection::intersectVoxel(const GeObject*& Obj,
				      const Vec3& Min, const Vec3& Max) const {
  VoxClass status1, status2;

  const GeObject* obj1;
  const GeObject* obj2;
  
  status1 = child1->intersectVoxel(obj1, Min, Max);
  status2 = child2->intersectVoxel(obj2, Min, Max);

  Obj = this;  // default return object is this CSG intersection
  
  // outside of one means outside of intersection
  if ((status1 == voxOutside) || (status2 == voxOutside)) {
    return voxOutside;
  }

  // both surfaces means we need to test against CSG combination
  if ((status1 == voxSurface) && (status2 == voxSurface)) {
    // there might be an opportunity for pruning here, in case a child ...
    // ... could simplify itself
    if ((child1 != obj1) || (child2 != obj2)) {
      // TODO: fix memory leak here ...

      Obj = new Intersection((GeObject*)obj1, (GeObject*)obj2);

      /*
      if (((GeObject*)obj1)->countPrimitives() +
	  ((GeObject*)obj2)->countPrimitives() != 2) {
	Obj = new Intersection((GeObject*)obj1, (GeObject*)obj2);
      } else {
	// special case when both children are primitive quadrics
	Obj = new QuadricIntersection((GeObject*)obj1, (GeObject*)obj2);
      }
      */

      if (Obj == NULL) {
	fprintf(stderr,
		"Intersection::intersectVoxel(): no memory for pruned CSG.\n");
	exit(0);
      }
    }
    return voxSurface;
  }

  // surface of only one means we can prune CSG tree
  if (status1 == voxSurface) {
    Obj = obj1;
    return voxSurface;
  }
  if (status2 == voxSurface) {
    Obj = obj2;
    return voxSurface;
  }

  // both inside means inside of intersection
  return voxInside;
}

double Intersection::intersectRay(Ray& Ray, double MinDist,
				 const GeObject*& Primitive) const {
  double t1,t2;
  
  const GeObject* hit1;  // the primitive actually hit
  const GeObject* hit2; 
  
  // global statistics: count CSG intersection tests
#if defined(STATISTICS)
  OSAtomicIncrement64(&numIntersection);
#endif
  
  // quick and dirty implementation
  int iter = 0;  // keep track of number of tries
  do {
#if defined(STATISTICS)
  OSAtomicIncrement64(&numIntersectionStep);
#endif
    iter++;
    t1 = child1->intersectRay(Ray, MinDist, hit1);
    t2 = child2->intersectRay(Ray, MinDist, hit2);
    
    if ((t1 == MAXNUM) && (t2 == MAXNUM)) { // bail out when no hit
      Primitive = this;
      return t1;  // return MAXNUM
    }
   
    // sort intersections; if near hit is valid, return it
    if (t1 < t2) {
      if (child2->evaluatePoint(Ray.evaluate(t1)) <= CSG_EPSILON_FUDGE) {
	Primitive = hit1;
	return t1;
      }
      MinDist = t1;  // if invalid, advance MinDist and recurse
    } else {
      if (child1->evaluatePoint(Ray.evaluate(t2)) <= CSG_EPSILON_FUDGE) {
	Primitive = hit2;
	return t2;
      }
      MinDist = t2;  // if invalid, advance MinDist and recurse
    }
  } while (iter < 100);

  Primitive = this;  // defensive programming; should be redundant
  return MAXNUM;
}

double Intersection::evaluatePoint(const Vec3& Point) const {
  double val1, val2;

  val1 = child1->evaluatePoint(Point);
  val2 = child2->evaluatePoint(Point);

  return qmax(val1, val2);
}
  
Vec3 Intersection::surfaceNormal(const Vec3& Point) const {
  double val1, val2;

  // quick and dirty child selection via function value

  val1 = child1->evaluatePoint(Point);
  val2 = child2->evaluatePoint(Point);

  val1 = fabs(val1);
  val2 = fabs(val2);

  if (val1 < val2) {
    return child1->surfaceNormal(Point);
  } else {
    return child2->surfaceNormal(Point);
  }
}

Union::Union(GeObject* Child1, GeObject* Child2) {
  child1 = Child1;
  child2 = Child2;
  numPrimitives = 0;  // clear cache
}

VoxClass Union::intersectVoxel(const GeObject*& Obj,
			       const Vec3& Min, const Vec3& Max) const {
  VoxClass status1, status2;

  const GeObject* obj1;
  const GeObject* obj2;
  
  status1 = child1->intersectVoxel(obj1, Min, Max);
  status2 = child2->intersectVoxel(obj2, Min, Max);

  Obj = this;  // default return object is this CSG union
  
  // inside of one means inside union
  if ((status1 == voxInside) || (status2 == voxInside))
    return voxInside;

  // both surfaces means we need to test against CSG combination
  if ((status1 == voxSurface) && (status2 == voxSurface)) {
    // there might be an opportunity for pruning here, in case a child ...
    // ... could simplify itself
    if ((child1 != obj1) || (child2 != obj2)) {
      // TODO: memory leak here ...
      Obj = new Union((GeObject*)obj1, (GeObject*)obj2);
      if (Obj == NULL) {
	fprintf(stderr,
		"Union::intersectVoxel(): no memory for pruned CSG.\n");
	exit(0);
      }
    }
    return voxSurface;
  }

  // surface of only one means we can prune CSG tree
  if (status1 == voxSurface) {
    Obj = obj1;
    return voxSurface;
  }
  if (status2 == voxSurface) {
    Obj = obj2;
    return voxSurface;
  }

  // outside of both means outside of union
  return voxOutside;
}

double Union::intersectRay(Ray& Ray, double MinDist,
			  const GeObject*& Primitive) const {
  double t1,t2;
  
  const GeObject* hit1;  // the primitive actually hit
  const GeObject* hit2; 
  
  // global statistics: count CSG union tests
#if defined(STATISTICS)
  OSAtomicIncrement64(&numUnion);
#endif
  
  // quick and dirty implementation
  int iter = 0;  // keep track of number of tries
  do {
    iter++;
    t1 = child1->intersectRay(Ray, MinDist, hit1);
    t2 = child2->intersectRay(Ray, MinDist, hit2);
    
    if ((t1 == MAXNUM) && (t2 == MAXNUM)) { // bail out when no hit
      Primitive = this;
      return t1;  // return MAXNUM
    }
   
    // sort intersections; if near hit is valid, return it
    if (t1 < t2) {
      if (child2->evaluatePoint(Ray.evaluate(t1)) >= -CSG_EPSILON_FUDGE) {
	Primitive = hit1;
	return t1;
      }
      MinDist = t1;  // if invalid, advance MinDist and recurse
    } else {
      if (child1->evaluatePoint(Ray.evaluate(t2)) >= -CSG_EPSILON_FUDGE) {
	Primitive = hit2;
	return t2;
      }
      MinDist = t2;  // if invalid, advance MinDist and recurse
    }
  } while (iter < 100);

  Primitive = this;  // defensive programming; should be redundant
  return MAXNUM;
}

double Union::evaluatePoint(const Vec3& Point) const {
  double val1, val2;

  val1 = child1->evaluatePoint(Point);
  val2 = child2->evaluatePoint(Point);

  return qmin(val1, val2);
}
  
Vec3 Union::surfaceNormal(const Vec3& Point) const {
  double val1, val2;

  // quick and dirty child selection via function value

  val1 = child1->evaluatePoint(Point);
  val2 = child2->evaluatePoint(Point);

  val1 = fabs(val1);
  val2 = fabs(val2);

  if (val1 < val2) {
    return child1->surfaceNormal(Point);
  } else {
    return child2->surfaceNormal(Point);
  }
}
