/*************************************************************************
 *                        Boar 's Only A Raytracer                       *
 *                                                                       *
 *       Copyright (C) 2014  Holger Bettag  (hobold@vectorizer.org)      *
 *                                                                       *
 * This program is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation, either version 3 of the License, or     *
 * (at your option) any later version.                                   *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *************************************************************************/

#ifndef GEOBJECT_H
#define GEOBJECT_H

// for NULL
#include <stdlib.h>

#include "calc.h"
#include "material.h"
#include "ray.h"
#include "transform.h"

// for voxel classification (voxel here means axis aligned box)
// interesting cases are: voxel is contained in GeObject, voxel does not
// overlap GeObject, or voxel may contain surface points of GeObject.
// (voxel can also contain GeObject, this counts as voxSurface)
enum VoxClass {voxSurface, voxOutside, voxInside};

// virtual base class of geometric objects (possibly invisible or composite)
class GeObject {
 protected:
  unsigned int objectNumber;
  
 public:
  inline GeObject(void) {
    objectNumber = 0;
  };
  
  // geometry related methods

  // determine overlap between GeObject and axis aligned box
  // (see definition of enum VoxClass)
  // also sets an object pointer, for pruning of CSG trees
  virtual VoxClass intersectVoxel(const GeObject*& Obj,
				  const Vec3& Min, const Vec3& Max) const = 0;

  // number of primitives (> 1 for a CSG subtree)
  virtual int countPrimitives(void) = 0;

  // affine transform (translate, scale, rotate, shear) an object
  virtual void applyTransform(const Transform& trans) = 0;

  // invert inside/outside relation (for CSG)
  virtual void insideOut(void) = 0;

  // expects unit length ray direction, returns ray parameter in case of
  // hit, MAXNUM in case of miss
  // hits in front of minDist must be ignored
  // returns Primitive that's actually hit for efficient CSG
  virtual double intersectRay(Ray& Ray, double MinDist,
			      const GeObject*& Primitive) const = 0;
  
  // determine relative location of point
  // <0 means inside, ==0 means on surface, >0 means outside
  // (some imprecision near surface allowed: actual implementations need
  // not consistently return zero or specified sign for surface points;
  // primitives based on implicit surfaces may simply evaluate the function
  // they are based on)
  virtual double evaluatePoint(const Vec3& Point) const = 0;
  
  // calculate (outward pointing) surface normal of unit length
  // (this routine may assume that Point is on surface)
  virtual Vec3 surfaceNormal(const Vec3& Point) const = 0;

  // material related methods

  // obtain material
  virtual Material* getMaterial(void) const = 0;
  
  // set material (apply to untextured children in case of composites)
  // by default, only sets material when pointer in instance is still NULL
  virtual void setMaterial(const Material* Mat,
			   bool forceReplace = false) = 0;
};


// virtual base class of primitive objects (visible, with material)
class PrimObject : public GeObject {
 public:
  PrimObject(void) {
    this->material = NULL;
  };

  // geometry related methods

  virtual VoxClass intersectVoxel(const GeObject*& Obj,
				  const Vec3& Min, const Vec3& Max) const = 0;

  virtual int countPrimitives(void);

  virtual void applyTransform(const Transform& trans) = 0;

  virtual void insideOut(void) = 0;

  virtual double intersectRay(Ray& Ray, double MinDist,
			      const GeObject*& Primitive) const = 0;
  
  virtual double evaluatePoint(const Vec3& Point) const = 0;
  
  virtual Vec3 surfaceNormal(const Vec3& Point) const = 0;

  // material related methods

  // obtain material
  virtual Material* getMaterial(void) const;
  
  // set material
  virtual void setMaterial(const Material* Mat, bool forceReplace = false);
  
 protected:
  Material* material;

};

#endif // GEOBJECT_H
