/*************************************************************************
 *                        Boar 's Only A Raytracer                       *
 *                                                                       *
 *       Copyright (C) 2014  Holger Bettag  (hobold@vectorizer.org)      *
 *                                                                       *
 * This program is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation, either version 3 of the License, or     *
 * (at your option) any later version.                                   *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *************************************************************************/

#include <assert.h>

#include "material.h"
#include "lightsource.h"
#include "ray.h"
#include "trace.h"

const int MaxDepth = 10;  // maximum ray recursion depth
const double SELFINTERSECT_FUDGE = 0.0005;
//#define SELFINTERSECT_FUDGE 0.003

Material::Material(Vec3 Pigment, double Diffuse, double Ambient,
		   double HighlightIntensity, double HighlightSize,
		   double Reflection)
{
  colour = Pigment;
  diffuse = Diffuse;
  ambient = Ambient;
  hiliteIntensity = HighlightIntensity;
  hiliteSize = 1.0/HighlightSize;  // save divides later
  reflection = Reflection;
}

Vec3 Material::lighting(Vec3 Dir, Vec3 Pos, Vec3 Normal, const Scene& theScene,
#if defined(ANTIALIAS)
			uintptr_t& Hash,
#endif  // defined(ANTIALIAS)
			int Depth) const {
  Vec3 result = colour;
  const ListNode* lightList = theScene.lights.firstNode();
  LightSource* curLight;
  double weight, cosAngle;
  Vec3 lightDir;     // direction vector towards light source
  Vec3 lightColour;
  Vec3 reflectedDir; // direction of specularly reflected ray
  double tmp;

  // find reflected ray direction
  tmp = 2.0*Normal.dot(Dir);
  reflectedDir.x = Dir.x - tmp*Normal.x;
  reflectedDir.y = Dir.y - tmp*Normal.y;
  reflectedDir.z = Dir.z - tmp*Normal.z;

#if defined (NORMALIZE)
  
  reflectedDir.normalize();
  
#else

  assert(reflectedDir.len() > 0.9999);
  assert(reflectedDir.len() < 1.0001);

#endif


  result.scale(ambient);  // start with ambient light level

  // to prevent self-shadowing, move Pos a tiny bit along Normal
  Pos.x += SELFINTERSECT_FUDGE * Normal.x;
  Pos.y += SELFINTERSECT_FUDGE * Normal.y;
  Pos.z += SELFINTERSECT_FUDGE * Normal.z;

  // diffuse reflection and highlights for each lightsource
  while (lightList != NULL) {
    curLight = (LightSource*)lightList->getData();
    weight = curLight->visibilityFrom(Pos, theScene);
    if (weight != 0.0) {    // only proceed if this light is visible

#if defined(ANTIALIAS)
      // add visible light sources to light path hash
      Hash ^= uintptr_t(curLight);
      Hash = (Hash << 17) | (Hash >> ((sizeof(uintptr_t) << 3) - 17));
#endif  // defined(ANTIALIAS)
      
      lightDir = curLight->directionFrom(Pos);
      cosAngle = lightDir.dot(Normal);
      if (cosAngle > 0.0) { // only proceed if object faces light

	lightColour = curLight->incomingLight(Pos);

	// diffuse light
	weight *= cosAngle;  // dim light according to angle of incidence
	weight *= diffuse;   // weight with diffuse reflectivity

	// filter light color with object pigment
	result.x += (weight*lightColour.x) * colour.x;
	result.y += (weight*lightColour.y) * colour.y;
	result.z += (weight*lightColour.z) * colour.z;

	// specular highlight

	// compare reflected ray direction to lightDir
	tmp = reflectedDir.dot(lightDir);
	
	if (tmp > 0.0) {  // only proceed for similar directions (<90 degrees)
	  
#if !defined(NORMALIZE)

	  assert(tmp <= 1.0001);

#endif

	  /*
	  if (tmp > 1.0)
	    tmp = 1.0;
	  */

	  // resize highlight (shrinks with increasing value)
	  tmp = ((tmp - 1.0) * hiliteSize) + 1.0;
	  
#if !defined(NORMALIZE)

	  assert(tmp <= 1.0008);  // triggered at 1.0007 by example huge res

#endif

	  if (tmp > 0.0) {  // only proceed in area of shrinked highlight

#if defined(ANTIALIAS)
	    // add visible highlights to light path hash
	    // for a bit of better-than-nothing highlight antialiasing
	    Hash ^= uintptr_t(curLight);
	    Hash = (Hash << 13) | (Hash >> ((sizeof(uintptr_t) << 3) - 13));
#endif  // defined(ANTIALIAS)

	    tmp *= tmp;

#if !defined(NORMALIZE)

	    assert(tmp <= 1.0015);  // triggered at 1.0014 by example huge res

#endif
	   
	    tmp *= hiliteIntensity;

	    // add colour of lightsource as highlight
	    result.x += tmp*lightColour.x;
	    result.y += tmp*lightColour.y;
	    result.z += tmp*lightColour.z;
	  }
	}
      }
    }
    lightList = lightList->nextNode();
  }

  // perfect specular reflection, but prevent infinite recursion
  if ((reflection != 0.0) && (Depth < MaxDepth)) {
    Ray ray;
    
    ray.base = Pos;          // the slightly moved position
    ray.dir = reflectedDir;

    lightColour = TraceRay(ray, theScene,
#if defined(ANTIALIAS)
			   Hash,
#endif  // defined(ANTIALIAS)
			   Depth + 1);

    result.x += reflection * lightColour.x;
    result.y += reflection * lightColour.y;
    result.z += reflection * lightColour.z;

  }

  return result;
}

