/*************************************************************************
 *                        Boar 's Only A Raytracer                       *
 *                                                                       *
 *       Copyright (C) 2014  Holger Bettag  (hobold@vectorizer.org)      *
 *                                                                       *
 * This program is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation, either version 3 of the License, or     *
 * (at your option) any later version.                                   *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include <assert.h>

#include "parser.h"

#include "quadric.h"
#include "csg.h"

// number of generated Quadrics, defined in quadric.c++
extern unsigned long numberOfQuadrics;

// Parent may be NULL for the root
// If File is NULL, it will be obtained from FileName with fopen()
ParserContext::ParserContext(ParserStack* Stack, char* FileName,
			     ParserContext* Parent, FILE* File,
			     int LineNum) {

  if (FileName == NULL) {
    fprintf(stderr,"No filename given for ParserContext\n");
    exit(EXIT_FAILURE);
  }
  fileName = strdup(FileName);  // copy file name
  if (fileName == NULL) {
    fprintf(stderr,"no memory for ParserContext file name\n");
    exit(EXIT_FAILURE);
  }
  
  if (File == NULL) {
    File = fopen(FileName, "r");
    if (File == NULL) {
      fprintf(stderr,"File \"%s\" not found in ParserContext\n", FileName);
      exit(EXIT_FAILURE);
    }
  }
  file = File;
  
  stack = Stack;
  parent = Parent;
  lineNum = LineNum;
}

ParserContext::~ParserContext() {
  if (file != NULL) {
    fclose(file);
  }
  
  free(fileName);  // delete copy of file name

  assert(stack != NULL);  // check for valid ParserStack

  // check for inconsistent ParserStack state
  assert(stack->topOfStack == this);
  
  // pop me off the stack
  stack->topOfStack = parent;
}

void ParserContext::printWarning(const char* const Message) const {
  fprintf(stderr, "Warning in file \"%s\" line %d (token %s):\n  %s\n",
	  fileName, lineNum, stack->token, Message);
}

void ParserContext::printError(const char* const Message) const {
  fprintf(stderr, "Error in file \"%s\" line %d (token %s):\n  %s\n",
	  fileName, lineNum, stack->token, Message);
}

void ParserContext::exitWithError(const char* const Message) const {
  printError(Message);
  exit(EXIT_FAILURE);
}

// check for special terminating chars
bool isTerminator(char c) {
  return (c == '<') || (c == '>')
    || (c == '{') || (c == '}');
}


// the basic idea is to skip all whitespace and comments before, but not
// after the token; so lineNum will automatically be accurate
char* ParserContext::nextToken(void) {
  int l = 0;   // token tength
  char c;      // current char
  char* buf = stack->token;  // shortcut to buffer

  buf[0] = 0;  // initialize to empty string

  // skip whitespace but count newlines
  do {
    c = getc(file);
    if (c == EOF) {
      return stack->token;  // string still empty
    }
    if (c == '\n') {
      lineNum++;
    }
  } while (isspace(c));

  // collect non-whitespace, non-terminating characters in buffer
  do {
    buf[l++] = c;
    
    // return terminating chars as standalone tokens
    if (isTerminator(c)) {
      if (l == 1) {
	// terminator is the only char so far, so just return it
	buf[1] = 0;
	return buf;
      }
      // terminator is preceded by other chars
      ungetc(c, file); // so stuff it back into stream
      l--;
      break;  // and terminate the loop
    }
    
    // process comment initiators as standalone tokens here, skip later
    if (l == 2) {
      if (buf[0] == '/') {
	if ((buf[1] == '/') || (buf[1] == '*')) {
	  // it is a comment, terminate the loop
	  break; 
	}
      }
    }
    
    // in other cases, a comment terminates a keyword
    if ((c == '/') && (l > 1)) {
      ungetc(c, file); // so stuff it back into stream
      l--;
      break;  // and terminate the loop
    }

    c = getc(file);
  } while ((!isspace(c)) && (c != EOF));
  buf[l] = 0;  // terminate string
  
  if (isspace(c)) {
    // so we don't forget to count this later in case it is a newline
    ungetc(c, file);
    // counting it now might lead to error messages with line numbers off by 1
  }

  // now we have a candidate for a token, but it could be a comment

  // check for comment delimiter
  if ((l < 2) || ((strncmp(buf, "//", 2) != 0)
      && (strncmp(buf, "/*", 2) != 0))) {
    // token is not a comment
    if ((strlen(stack->token) != 8)
	|| (strcasecmp(stack->token, "#include") != 0)) {
      // token is not an #include directive, so return it
      return stack->token;
    }
    // found an #include directive, so we need to set up a new context
    {
      // parse name of file to be #include'd
      
      do {  // skip whitespace, count newlines, handle EOF
	c = getc(file);
	if (c == EOF) {
	  exitWithError("end of file after #include");
	}
	if (c == '\n') {
	  lineNum++;
	}
      } while (isspace(c));

      if (c != '"') {
	exitWithError("filename given for #include must be enclosed "
		      "in double quotes");
      }
      
      l = 0;  // reset string length of file name to be parsed
      
      do {
	c = getc(file);
	if (c == EOF) {
	  exitWithError("end of file in #include file name");
	}
	buf[l++] = c;
      } while (c != '"');
      
      buf[l-1] = 0;  // terminate string, overwriting closing double quote

      char* newFileName = strdup(buf);  // copy file name string
      ParserContext* newContext = new ParserContext(stack, newFileName, this);

      if (newContext == NULL) {
	exitWithError("out of memory for ParserContext");
      }
      stack->topOfStack = newContext;  // push on ParserStack
      
      return newContext->nextToken();  // fetch token from new context
      /*
      fprintf(stderr,"token after #include: %s\n", stack->token);
      return stack->token;
      */
    }
  }

  // otherwise skip comment, then recurse
  if (buf[1] == '/') {
    // double slash comment, skip until newline
    do {
      c = getc(file);
    } while ((c != '\n') && (c != EOF));
    lineNum++;  // wrong line numbers don't matter in case of unexpected EOF
  } else {
    // slash asterisk comment, skip until asterisk slash
    do {
      c = getc(file);
      if (c == '*') {  // check for asterisk slash delimiter
	c = getc(file);
	if (c == '/') {
	  break;  // exit loop in case of asterisk slash
	} else if (c == '*'){
	  // otherwise rewind the non-slash in case it is an asterisk
	  ungetc(c, file);
	}
      }
      if (c == EOF) {  // prevent infinite loop on EOF
	printError("end of file in multiline comment");
	break;
      }
      if (c == '\n')  // keep track of line number
	lineNum++;
    } while (true);
  }
  
  // leave tail recursion optmisation to the compiler
  return nextToken();
}


ParserStack::ParserStack(char* FileName) {

  token[0] = 0;     // initialize strings to empty
  message[0] = 0;

  // weak step to prevent buffer overruns
  token[1023] = 0;  // guard elements
  message[1023] = 0;
  
  topOfStack = new ParserContext(this, FileName);
  
  if (topOfStack == NULL) {
    fprintf(stderr, "No memory for ParserContext\n");
    exit(EXIT_FAILURE);
  }
}

ParserStack::~ParserStack() {
  // clean up ParserContexts if there are any
  while (topOfStack != NULL) {
    delete topOfStack;  // this will pop one context
  }
}

char* ParserStack::nextToken(void) {
  char* Token = topOfStack->nextToken();

  if (*Token == 0) {  // empty string signals EOF
    // if there is a parent context on the stack, continue parsing from there
    ParserContext* Parent = topOfStack->parent;
    if (Parent != NULL) {
      delete topOfStack;    // automatically pops itself off the stack
      return nextToken();   // simply recurse
    }
  }
  
  return Token;
}

// check for the presence of a particular token and consume it
void ParserStack::consumeLiteral(const char* Word) {
  if (strcasecmp(nextToken(), Word)) {
    exitWithError("literal expected");
  }
  return;
}

double ParserStack::consumeFloat(void) {
  double result;
  
  if (1 != sscanf(nextToken(), "%lf", &result)) {
    exitWithError("float expected");
  }

  return result;
}

Vec3 ParserStack::consumeVector(void) {
  Vec3 result;

  consumeLiteral("<");
  result.x = consumeFloat();
  result.y = consumeFloat();
  result.z = consumeFloat();
  consumeLiteral(">");

  return result;
}

// TagList parser
// tag lists are bracketed by '{' and '}'
void ParserStack::consumeTags(TagList& Tags) {
  char* tok;
  int i;

  // clean up TagItems
  for (i = Tags.size - 1; i >= 0; i--) {
    Tags.item[i].isSet = false;
  }

  consumeLiteral("{");  // check for opening bracket

  do {
    tok = nextToken();
   
    if (tok[0] == 0) {
      exitWithError("encountered end of file before '}'");
    }

    if (strcmp(tok, "}") == 0)
      break;    // closing bracket terminates loop

    // search taglist for matching keyword
    i = Tags.size - 1;
    while ((i >= 0) && (strcasecmp(tok, Tags.item[i].name) != 0)) {
      i = i - 1;
    }
    
    if (i < 0) {  // not found?
      exitWithError("unknown tag name");
    }
    
    // parse this tag item (at index i)
    if (Tags.item[i].isSet == true) {  // multiple tag definitions?
      exitWithError("duplicate tag");
    }

    Tags.item[i].isSet = true;

    switch (Tags.item[i].type) {
    case FloatType:
      // read float
      Tags.item[i].data.x = consumeFloat();
      break;
      
    case VectorType:
      // read vector
      Tags.item[i].data = consumeVector();
      break;

    case BoolType:
      // no value needs to be parsed
      break;
      
    default:
      exitWithError("unknown tag type");
    }

  } while (true);  // next tag
}

void ParserStack::consumeMaterial(GeObject* object) {
  Vec3 pigment;
  double ambient, diffuse, highlight, roughness, reflection;

  TagItem items[6] = {
    {"pigment",    VectorType, false, Vec3(0.9,  0.1, 0.1) },
    {"ambient",    FloatType,  false, Vec3(0.2,  0.0, 0.0) },
    {"diffuse",    FloatType,  false, Vec3(0.8,  0.0, 0.0) },
    {"highlight",  FloatType,  false, Vec3(0.6,  0.0, 0.0) },
    //{"highlight",  FloatType,  false, Vec3(0.0,  0.0, 0.0) },
    {"roughness",  FloatType,  false, Vec3(0.1, 0.0, 0.0) },
    {"reflection", FloatType,  false, Vec3(0.0,  0.0, 0.0) },
  };
  TagList tags = {6, items};

  consumeTags(tags);
  
  pigment = items[0].data;
  ambient = items[1].data.x;
  diffuse = items[2].data.x;
  highlight = items[3].data.x;
  roughness = items[4].data.x;
  reflection = items[5].data.x;

  Material* mat = new Material(pigment, diffuse, ambient, highlight,
			       roughness, reflection);
  if (mat == NULL) {
    exitWithError("no memory for Material");
  }

  object->setMaterial(mat);
}

void ParserStack::consumeScale(Transform& trans) {
  Vec3 factors = consumeVector();
  trans.Scale(factors);
}


void ParserStack::consumeTranslate(Transform& trans) {
  Vec3 offset = consumeVector();
  trans.Translate(offset);
}


void ParserStack::consumeRotateX(Transform& trans) {
  double angle = consumeFloat();
  trans.RotateX(angle);
}


void ParserStack::consumeRotateY(Transform& trans) {
  double angle = consumeFloat();
  trans.RotateY(angle);
}


void ParserStack::consumeRotateZ(Transform& trans) {
  double angle = consumeFloat();
  trans.RotateZ(angle);
}

void ParserStack::consumeRotate(Transform& trans) {
  Vec3 angles = consumeVector();
  trans.RotateX(angles.x);
  trans.RotateY(angles.y);
  trans.RotateZ(angles.z);
}

void ParserStack::consumeTransform(Transform& trans) {
  Vec3 c1 = consumeVector();
  Vec3 c2 = consumeVector();
  Vec3 c3 = consumeVector();
  
  Transform temp = Transform(c1, c2, c3, Vec3(0.0, 0.0, 0.0));

  trans.Apply(temp);
}

bool ParserStack::tryTransform(Transform& trans) {
  bool isTransform = true;
  bool transformed = false;
  char* buf = currentToken();

  // accumulate all successive transforms
  while(isTransform) {
    if (strcasecmp(buf, "Scale") == 0) {
      consumeScale(trans);
      transformed = true;
      buf = nextToken();
    } else if (strcasecmp(buf, "Translate") == 0) {
      consumeTranslate(trans);
      transformed = true;
      buf = nextToken();
    } else if (strcasecmp(buf, "RotateX") == 0) {
      consumeRotateX(trans);
      transformed = true;
      buf = nextToken();
    } else if (strcasecmp(buf, "RotateY") == 0) {
      consumeRotateY(trans);
      transformed = true;
      buf = nextToken();
    } else if (strcasecmp(buf, "RotateZ") == 0) {
      consumeRotateZ(trans);
      transformed = true;
      buf = nextToken();
    } else if (strcasecmp(buf, "Rotate") == 0) {
      consumeRotate(trans);
      transformed = true;
      buf = nextToken();
    } else if (strcasecmp(buf, "Transform") == 0) {
      consumeTransform(trans);
      transformed = true;
      buf = nextToken();
    } else {
      isTransform = false;
    }
  }

  // return true if we ever encountered a transform
  return transformed;
}

void ParserStack::consumeObjectAttributes(GeObject* object) {
  char* buf;
  Transform trans;      // default constructor initializes identity transform
  bool seenMaterial = false;

  do {
    tryTransform(trans);
    buf = currentToken();

    if (strcmp(buf, "}") == 0)
      break;    // closing bracket terminates loop

    if (strcasecmp(buf, "material") == 0) {
      if(seenMaterial)
	exitWithError("duplicate Material");
      seenMaterial = true;
      consumeMaterial(object);
    } else {
      exitWithError("unrecognized keyword in object attributes");
    }
    buf = nextToken();
    if (buf[0] == 0)
      exitWithError("unexpected end of file in object attributes");
  } while (true);

  // move the object according to the accumulated transform
  object->applyTransform(trans);
}

bool ParserStack::tryInvert(bool invertMe) {
  char* buf = nextToken();
  if (strcasecmp(buf, "Invert") == 0) {
    // if token equals "Invert", toggle flag
    invertMe = !invertMe;
    buf = nextToken();  // consume keyword
  }
  return invertMe;
}

GeObject* ParserStack::consumeGenericObject(bool invertMe) {
  GeObject* object;

  // check for opening bracket
  consumeLiteral("{");

  invertMe = tryInvert(invertMe);

  object = tryObject(invertMe);
  if (object == NULL)
    exitWithError("expected geometric primitive");

  nextToken();
  consumeObjectAttributes(object);

  return object;
}

GeObject* ParserStack::consumeSphere(bool invertMe) {
  // check for opening bracket
  consumeLiteral("{");

  Vec3 center = consumeVector();
  double radius = consumeFloat();

  invertMe = tryInvert(invertMe);

  Quadric* object = new Quadric;

  if (object == NULL) {
    exitWithError("no memory for Sphere");
  }
  
  object->initSphere(center, radius);

  if (invertMe == true)
    object->insideOut();

  consumeObjectAttributes(object);

  return object;
}

GeObject* ParserStack::consumeCylinder(bool invertMe, bool capped) {
  GeObject* object;

  // check for opening bracket
  consumeLiteral("{");

  Vec3 point1 = consumeVector();
  Vec3 point2 = consumeVector();
  double radius = consumeFloat();
  
  invertMe = tryInvert(invertMe);

  Quadric* tube = new Quadric;
  
  if (tube == NULL) {
    exitWithError("no memory for Cylinder");
  }

  tube->initCylinder(point1, point2, radius);
  if (invertMe == true)
    tube->insideOut();

  if (capped == false) {
    object = tube;
  } else {
    Quadric* slab = new Quadric;

    if (slab == NULL) {
      exitWithError("no memory for Cylinder caps");
    }
    slab->initSlab(point1, point2);
    if (invertMe == true)
      slab->insideOut();

    if (invertMe == false) {
      object = new Intersection(tube, slab);
    } else {
      object = new Union(tube, slab);
    }
  }

  consumeObjectAttributes(object);
  
  return object;
}

GeObject* ParserStack::consumeCone(bool invertMe, bool capped) {
  GeObject* object;

  // check for opening bracket
  consumeLiteral("{");

  Vec3 point1 = consumeVector();
  double radius1 = consumeFloat();
  Vec3 point2 = consumeVector();
  double radius2 = consumeFloat();

  invertMe = tryInvert(invertMe);

  Quadric* cone = new Quadric;
  
  if (cone == NULL) {
    exitWithError("no memory for Cone");
  }

  cone->initCone(point1, point2, radius1, radius2);
  if (invertMe == true)
    cone->insideOut();

  if (capped == false) {
    object = cone;
  } else {
    Quadric* slab = new Quadric;
    if (slab == NULL) {
      exitWithError("no memory for Cone caps");
    }
    slab->initSlab(point1, point2);
    if (invertMe == true)
      slab->insideOut();

    if (invertMe == false) {
      object = new Intersection(cone, slab);
    } else {
      object = new Union(cone, slab);
    }
  }

  consumeObjectAttributes(object);

  return object;
}

GeObject* ParserStack::consumeSlab(bool invertMe) {
  // check for opening bracket
  consumeLiteral("{");

  Vec3 point1 = consumeVector();
  Vec3 point2 = consumeVector();
  
  invertMe = tryInvert(invertMe);

  Quadric* slab = new Quadric;
  
  if (slab == NULL) {
    exitWithError("no memory for Slab");
  }
  
  slab->initSlab(point1, point2);
  if (invertMe == true)
    slab->insideOut();

  consumeObjectAttributes(slab);

  return slab;
}

GeObject* ParserStack::consumePlanes(bool invertMe) {
  // check for opening bracket
  consumeLiteral("{");

  Vec3 point1 = consumeVector();
  Vec3 normal1 = consumeVector();
  if (normal1.len() == 0.0)
    exitWithError("plane normal cannot be zero vector");

  Vec3 point2 = consumeVector();
  Vec3 normal2 = consumeVector();
  if (normal2.len() == 0.0)
    exitWithError("plane normal cannot be zero vector");

  invertMe = tryInvert(invertMe);

  Quadric* planes = new Quadric;
  
  if (planes == NULL) {
    exitWithError("no memory for Planes");
  }

  planes->initPlanes(normal1, -normal1.dot(point1),
		     normal2, -normal2.dot(point2));
  
  if (invertMe == true) {
    planes->insideOut();
  }

  consumeObjectAttributes(planes);

  return planes;
}

GeObject* ParserStack::consumeQuadric(bool invertMe) {
  // check for opening bracket
  consumeLiteral("{");

  Vec3 squared = consumeVector();
  Vec3 mixed = consumeVector();
  Vec3 linear = consumeVector();
  double offset = consumeFloat();
  
  invertMe = tryInvert(invertMe);

  Quadric* quadric = new Quadric(squared, mixed, linear, offset);
  
  if (quadric == NULL) {
    exitWithError("no memory for Quadric");
  }
  
  if (invertMe == true)
    quadric->insideOut();

  consumeObjectAttributes(quadric);

  return quadric;
}


GeObject* ParserStack::consumeLathePiece(bool invertMe, bool capped) {
  GeObject* object;

  // check for opening bracket
  consumeLiteral("{");

  consumeLiteral("<");
  double u1 = consumeFloat();
  double v1 = consumeFloat();
  consumeLiteral(">");
  consumeLiteral("<");
  double u2 = consumeFloat();
  double v2 = consumeFloat();
  consumeLiteral(">");
  double d = consumeFloat();

  invertMe = tryInvert(invertMe);
  
  d = d + d;

  double a = d*(d + 2*(u2 + u1)) - (u2 - u1)*(u2 - u1);
  double b = 2*(u2-u1)*(u2*v1 - u1*v2) - d*(d*(v1+v2) + 2*(u2 + u1)*(v1 + v2));
  double c = d*v1*v2*(d + 2*(u2 + u1)) - (u2*v1 - u1*v2)*(u2*v1 - u1*v2);

  double den = (v1-v2);
  den = 1.0/(den*den);

  a *= den;
  b *= den;
  c *= den;

  Quadric* lathe = new Quadric(Vec3(1.0, a, 1.0),
			       Vec3(0, 0, 0),
			       Vec3(0, b, 0), c);

  if (lathe == NULL) {
    exitWithError("no memory for Lathe Piece");
  }

  if (invertMe == true)
    lathe->insideOut();

  if (capped == false) {
    object = lathe;
  } else {
    Quadric* slab = new Quadric;
    if (slab == NULL) {
      exitWithError("no memory for Cone caps");
    }
    slab->initSlab(Vec3(0.0, v1, 0.0), Vec3(0.0, v2, 0.0));
    if (invertMe == true)
      slab->insideOut();

    if (invertMe == false) {
      object = new Intersection(lathe, slab);
    } else {
      object = new Union(lathe, slab);
    }
  }

  consumeObjectAttributes(object);

  return object;
}

Quadric* ParserStack::consumeQuadricObject(bool invertMe) {
  nextToken();
  Quadric* quadric = tryQuadricObject(invertMe);
  
  if (quadric == NULL)
    exitWithError("expected quadric primitive");

  return quadric;
}

Quadric* ParserStack::tryQuadricObject(bool invertMe) {
  Quadric* quadric = NULL;
  char* buf = currentToken();

  if (buf[0] == 0) {
    exitWithError("unexpected end of file in quadric primitive");
  }

  if (strcasecmp(buf, "Sphere") == 0) {
    quadric = (Quadric*)consumeSphere(invertMe);
    
  } else if (strcasecmp(buf, "Cylinder") == 0) {
    quadric = (Quadric*)consumeCylinder(invertMe);
    
  } else if (strcasecmp(buf, "Cone") == 0) {
    quadric = (Quadric*)consumeCone(invertMe);
    
  } else if (strcasecmp(buf, "Slab") == 0) {
    quadric = (Quadric*)consumeSlab(invertMe);
    
  } else if (strcasecmp(buf, "Planes") == 0) {
    quadric = (Quadric*)consumePlanes(invertMe);
    
  } else if (strcasecmp(buf, "Quadric") == 0) {
    quadric = (Quadric*)consumeQuadric(invertMe);
    
  } else if (strcasecmp(buf, "LathePiece") == 0) {
    quadric = (Quadric*)consumeLathePiece(invertMe);
  }
  
  // return NULL in case of no match
  return quadric;
}

GeObject* ParserStack::tryObject(bool invertMe) {
  GeObject* object;
  char* buf = currentToken();

  if (buf[0] == 0) {
    exitWithError("unexpected end of file in object");
  }

  object = tryQuadricObject(invertMe);
  if (object != NULL)
    return object;

  if (strcasecmp(buf, "Union") == 0) {
    object = consumeCSG(false, invertMe);
    
  } else if (strcasecmp(buf, "Intersection") == 0) {
    object = consumeCSG(true, invertMe);
    
  } else if (strcasecmp(buf, "Difference") == 0) {
    object = consumeDifference(invertMe);
    
  } else if (strcasecmp(buf, "Box") == 0) {
    object = consumeBox(invertMe);
    
  } else if (strcasecmp(buf, "Object") == 0) {
    object = consumeGenericObject(invertMe);
    
  } else if (strcasecmp(buf, "CappedCylinder") == 0) {
    object = consumeCylinder(invertMe, true);
    
  } else if (strcasecmp(buf, "CappedCone") == 0) {
    object = consumeCone(invertMe, true);
    
  } else if (strcasecmp(buf, "CappedLathePiece") == 0) {
    object = consumeLathePiece(invertMe, true); 
  }
  
  // return NULL in case of no match
  return object;
}

GeObject* ParserStack::consumeCSG(bool isIntersection, bool invertMe) {
  char* buf;
  GeObject* object;
  GeObject* composite;
  Transform trans;      // default constructor initializes identity transform
  int size = 4, num = 0;  // array size, number of occupied entries
  int i;
  GeObject** objects = (GeObject**)malloc(size * sizeof(GeObject*));

  // check for opening bracket
  consumeLiteral("{");

  invertMe = tryInvert(invertMe);

  do {  // collect objects in array, growing array if necessary
    object = tryObject(invertMe);
    if (object == NULL) {
      // no object matched, terminate loop
      break;
    }
    buf = nextToken();

    // insert object into array
    if (num >= size) {  // still room in array?
      // no room, enlarge array first
      size = size + size;
      objects = (GeObject**)realloc(objects, size * sizeof(GeObject*));
      if (objects == NULL) {
	exitWithError("out of memory in CSG");
      }
    }
    // everything ok, insert object
    objects[num++] = object;

    // is there a transform specified inbetween the objects?
    if (tryTransform(trans)) {
      // if there is a valid transform, apply it to all objects seen so far
      for (i = 0; i < num; i++) {
	objects[i]->applyTransform(trans);
      }
      trans = Transform();  // re-initialize with identity transform
    }
  } while (true);

  // combine objects with respect to CSG type
  if (isIntersection) {
    // individual objects have already been inverted accordingly, so
    // this will only change CSG type later on
    invertMe = !invertMe;
  }

  if (num == 0) {
    exitWithError("no objects in CSG");
  }

  if (num == 1) {
    printWarning("degenerate CSG with a single subobject");  // not fatal
  }

  while (num > 1) {
    // combine pairs of GeObjects into composites
    // each pass roughly halves the number of GeObjects
    for (i = 0; i < (num-1); i += 1) {
      if (invertMe == false) {
	composite = new Union(objects[i], objects[i+1]);
      } else {

	composite = new Intersection(objects[i], objects[i+1]);

	/*
	// are both children primitive quadrics?
	if (objects[i]->countPrimitives() +
	    objects[i+1]->countPrimitives() != 2) {
	  
	  composite = new Intersection(objects[i], objects[i+1]);
	} else {
	  // if both are primitive, use special case
	  composite = new QuadricIntersection(objects[i], objects[i+1]);
	}
	*/
      }
      if (composite == NULL)
	exitWithError("no memory for CSG object");

      num -= 1;  // number of objects has been decreased by one
      objects[i] = composite;
      objects[i+1] = objects[num];  // move last valid object into freed entry
    }
  }
  
  composite = objects[0];  // keep root of CSG tree
  free(objects);  // deallocate temporary array of objects

  consumeObjectAttributes(composite);

  return composite;
}

GeObject* ParserStack::consumeDifference(bool invertMe) {
  char* buf;
  GeObject* object;
  GeObject* composite;
  Transform trans;      // default constructor initializes identity transform
  int size = 4, num = 0;  // array size, number of occupied entries
  int i;
  GeObject** objects = (GeObject**)malloc(size * sizeof(GeObject*));

  // check for opening bracket
  consumeLiteral("{");

  invertMe = tryInvert(invertMe);

  do {  // collect objects in array, growing array if necessary
    object = tryObject(invertMe);
    if (object == NULL) {
      // no object matched, terminate loop
      break;
    }
    buf = nextToken();

    // insert object into array
    if (num >= size) {  // still room in array?
      // no room, enlarge array first
      size = size + size;
      objects = (GeObject**)realloc(objects, size * sizeof(GeObject*));
      if (objects == NULL) {
	exitWithError("out of memory in Difference");
      }
    }
    // everything ok, insert object
    objects[num++] = object;

    if (num == 1)   // trigger once after first object
      invertMe = !invertMe;  // remaining objects will be 'subtracted'

    // is there a transform specified inbetween the objects?
    if (tryTransform(trans)) {
      // if there is a valid transform, apply it to all objects seen so far
      for (i = 0; i < num; i++) {
	objects[i]->applyTransform(trans);
      }
      trans = Transform();  // re-initialize with identity transform
    }
  } while (true);

  if (num == 0)
    exitWithError("no objects in Difference");

  if (num == 1)
    exitWithError("degenerate Difference with a single subobject");

  invertMe = !invertMe;  // return to original sense of first object

  while (num > 2) {
    // combine pairs of GeObjects into composites, skipping first object
    // each pass roughly halves the number of GeObjects (after first)
    for (i = 1; i < (num-1); i += 1) {
      if (invertMe == true) {
	composite = new Union(objects[i], objects[i+1]);
      } else {
	composite = new Intersection(objects[i], objects[i+1]);
      }
      if (composite == NULL)
	exitWithError("out of memory for Difference objects");

      num -= 1;  // number of objects has been decreased by one
      objects[i] = composite;
      objects[i+1] = objects[num];  // move last valid object into freed entry
    }
  }

  // perform the 'subtraction'
  if (invertMe == true) {
    composite = new Union(objects[0], objects[1]);
  } else {
    composite = new Intersection(objects[0], objects[1]);
  }
  if (composite == NULL)
    exitWithError("out of memory for Difference object");
  
  free(objects);  // deallocate temporary array of objects

  consumeObjectAttributes(composite);

  return composite;
}

GeObject* ParserStack::consumeBox(bool invertMe) {
  Quadric* slab1;
  Quadric* slab2;
  GeObject* composite;

  // check for opening bracket
  consumeLiteral("{");

  Vec3 min = consumeVector();
  Vec3 max = consumeVector();

  invertMe = tryInvert(invertMe);
  
  slab1 = new Quadric;

  if (slab1 == NULL) {
    exitWithError("out of memory for box");
  }
  slab2 = new Quadric;
  
  if (slab2 == NULL) {
    exitWithError("out of memory for box");
  }
  
  slab1->initSlab(Vec3(min.x, 0.0, 0.0), Vec3(max.x, 0.0, 0.0));
  slab2->initSlab(Vec3(0.0, min.y, 0.0), Vec3(0.0, max.y, 0.0));

  if (invertMe == true) {
    slab1->insideOut();
    slab2->insideOut();
  }

  if (invertMe == true) {
    composite = new Union(slab1, slab2);
  } else {
    composite = new Intersection(slab1, slab2);
  }
  if (composite == NULL) {
    exitWithError("out of memory for box");
  }
  
  slab1 = new Quadric;
  
  if (slab1 == NULL) {
    exitWithError("out of memory for box");
  }

  slab1->initSlab(Vec3(0.0, 0.0, min.z), Vec3(0.0, 0.0, max.z));

  if (invertMe == true) {
    slab1->insideOut();
  }
  
  if (invertMe == true) {
    composite = new Union(slab1, composite);
  } else {
    composite = new Intersection(slab1, composite);
  }
  if (composite == NULL) {
    exitWithError("out of memory for box");
  }

  consumeObjectAttributes(composite);

  return composite;
}

GeObject* ParserStack::consumeObject(bool invertMe) {
  nextToken();
  GeObject* object = tryObject(invertMe);
  
  if (object == NULL)
    exitWithError("expected geometric object");

  return object;
}

LightSource* ParserStack::consumePointLight() {
  consumeLiteral("{");
  Vec3 position = consumeVector();
  Vec3 colour = consumeVector();
  consumeLiteral("}");

  PointLight* light = new PointLight(position, colour);
  
  if (light == NULL) {
    exitWithError("out of memory for PointLight");
  }

  return light;
}

Camera* ParserStack::consumeCamera() {
  Vec3 Eye, LookAt, Sky;
  double Width, Height, Distance;

  TagItem items[6] = {
    {"Eye",      VectorType, false, Vec3(0.0,  0.0, 10.0) },
    {"LookAt",   VectorType, false, Vec3(0.0,  0.0, 0.0) },
    {"Sky",      VectorType, false, Vec3(0.0,  1.0, 0.0) },
    {"Width",    FloatType,  false, Vec3(4.0,  0.0, 0.0) },
    {"Height",   FloatType,  false, Vec3(3.0,  0.0, 0.0) },
    {"Distance", FloatType,  false, Vec3(2.0,  0.0, 0.0) },
  };
  TagList tags = {6, items};
  
  consumeTags(tags);
  
  Eye = items[0].data;
  LookAt = items[1].data;
  Sky = items[2].data;
  Width = items[3].data.x;
  Height = items[4].data.x;
  Distance = items[5].data.x;

  PerspectiveCam* cam = new PerspectiveCam;
  if (cam == NULL) {
    exitWithError("out of memory for Camera");
  }

  cam->InitLookAt(Eye, LookAt, Sky, Width, Height, Distance);

  return cam;
}

// parse a scene file
Scene* ParserStack::parseScene(void) {
  char* buf;
  Camera* theCam;
  GeObject* object;
  LightSource* light;
  Scene* theScene = new Scene;

  static Material defaultMaterial =
    Material(Vec3(0.5, 0.5, 0.5),     // pigment
	     0.8,                     // diffuse
	     0.2,                     // ambient
	     //0.3,                     // highlight
	     0.0,                     // highlight
	     0.3,                     // roughness
	     0.0);                    // reflection 

  fprintf(stderr,"parsing\n");

  do {
    buf = nextToken();  // read next token
    if(buf[0] == 0) {
      break;  // terminate loop on EOF
    }

    object = tryObject();  // if token refers to a GeObject then parse it

    if (object != NULL) {
      object->setMaterial(&defaultMaterial);  // apply default if none is set
      theScene->objects.addItem(object);  // add to scene
    } else {
      if (strcasecmp(buf, "pointlight") == 0) {
	light = consumePointLight();
	theScene->lights.addItem(light);  // add to scene
      } else if (strcasecmp(buf, "camera") == 0) {
	theCam = consumeCamera();
	if (theScene->camera == NULL) {
	  theScene->camera = theCam;
	} else {
	  exitWithError("no more than one camera may be specified");
	}
      } else {
	exitWithError("unrecognized keyword in scene");
      }
    }
  } while (true);
  

  fprintf(stderr,"finished (%ld quadrics)\n", numberOfQuadrics);

  return theScene;
}
