/*************************************************************************
 *                        Boar 's Only A Raytracer                       *
 *                                                                       *
 *       Copyright (C) 2014  Holger Bettag  (hobold@vectorizer.org)      *
 *                                                                       *
 * This program is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation, either version 3 of the License, or     *
 * (at your option) any later version.                                   *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *************************************************************************/

#ifndef PARSER_H
#define PARSER_H

#include <stdio.h>

#include "calc.h"
#include "scene.h"
#include "geobject.h"
#include "transform.h"
#include "lightsource.h"
#include "camera.h"
#include "quadric.h"

// description of tag-value-pairs for parsing
enum DataType {FloatType, VectorType, BoolType};

struct TagItem {
  const char* name;  // parse this word
  DataType type;     // parse this data type
  bool isSet;        // has this tag been encountered during parsing?
  Vec3 data;         // data (scalar or vector)
};

struct TagList {
  int size;          // number of tags to parse
  TagItem* item;     // array of TagItems
};

class ParserStack;   // name declaration

class ParserContext {
 public:
  // Parent may be NULL for the root
  // If File is NULL, it will be obtained from FileName with fopen()
  ParserContext(ParserStack* Stack, char* FileName,
		ParserContext* Parent = NULL, FILE* File = NULL,
		int LineNum = 1);

  ~ParserContext();

  void printWarning(const char* const Message) const;

  void printError(const char* const Message) const;

  void exitWithError(const char* const Message) const;

  char* nextToken(void);

  ParserStack* stack;    // to locate the 'container'
  ParserContext* parent; // the 'return address'
  FILE* file;            // file being parsed
  char* fileName;        // that file's name
  int lineNum;           // current line number
};

class ParserStack {
 public:
  ParserStack(char* FileName);
  ~ParserStack();

  
  inline void printError(char* Message) {
    topOfStack->printError(Message);
  };

  inline void printWarning(char* Message) {
    topOfStack->printWarning(Message);
  };

  inline void exitWithError(char* Message) {
    topOfStack->exitWithError(Message);
  };

  inline char* currentToken(void) {
    return token;
  };
  
  char* nextToken(void);

  // parse a whole scene
  Scene* parseScene(void);

  ParserContext* topOfStack;  // the currently active ParserContext
  char token[1024];           // buffer for current token
  char message[1024];         // buffer for (error) messages

 protected:
  // simple parse functions, all consuming the next token
  void consumeLiteral(const char* Word);
  double consumeFloat(void);
  Vec3 consumeVector(void);

  // TagList parser
  // tag lists are bracketed by '{' and '}'
  void consumeTags(TagList& Tags);

  // specialized parser functions
  void consumeMaterial(GeObject* object);

  void consumeScale(Transform& trans);
  void consumeTranslate(Transform& trans);  
  void consumeRotateX(Transform& trans);
  void consumeRotateY(Transform& trans);
  void consumeRotateZ(Transform& trans);
  void consumeRotate(Transform& trans);
  void consumeTransform(Transform& trans);
  
  bool tryTransform(Transform& trans);
  bool tryInvert(bool invertMe);

  void consumeObjectAttributes(GeObject* object);

  GeObject* consumeGenericObject(bool invertMe = false);

  GeObject* consumeSphere(bool invertMe = false);
  GeObject* consumeCylinder(bool invertMe = false, bool capped = false);
  GeObject* consumeCone(bool invertMe = false, bool capped = false);
  GeObject* consumeSlab(bool invertMe = false);
  GeObject* consumePlanes(bool invertMe = false);
  GeObject* consumeQuadric(bool invertMe = false);
  GeObject* consumeLathePiece(bool invertMe = false, bool capped = false);
  Quadric* consumeQuadricObject(bool invertMe = false);

  GeObject* consumeCSG(bool isIntersection, bool invertMe = false);
  GeObject* consumeUnion(bool invertMe = false);
  GeObject* consumeIntersection(bool invertMe = false);
  GeObject* consumeDifference(bool invertMe = false);
  GeObject* consumeBox(bool invertMe = false);
  GeObject* consumeObject(bool invertMe = false);

  Quadric* tryQuadricObject(bool invertMe = false);
  GeObject* tryObject(bool invertMe = false);
  
  LightSource* consumePointLight(void);
  
  Camera* consumeCamera(void);
};

#endif // PARSER_H
