/*************************************************************************
 *                        Boar 's Only A Raytracer                       *
 *                                                                       *
 *       Copyright (C) 2014  Holger Bettag  (hobold@vectorizer.org)      *
 *                                                                       *
 * This program is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation, either version 3 of the License, or     *
 * (at your option) any later version.                                   *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *************************************************************************/

#include <assert.h>
#include <stdio.h>
#include <math.h>

#if defined(STATISTICS)
#include <libkern/OSAtomic.h>
#endif

#include "quadric.h"

#if defined(STATISTICS)
// global statistics counters, defined in boar.c++
extern int64_t numQuadric;
#endif

// number of generated quadrics, for parser statistics
unsigned long numberOfQuadrics = 0;

void Quadric::init(const Vec3& Squared, const Vec3& Mixed, const Vec3& Linear,
		   double Offset) {
  // store basic coefs
  A = Squared.x;
  B = Squared.y;
  C = Squared.z;
  D = Mixed.x;
  E = Mixed.y;
  F = Mixed.z;
  G = Linear.x;
  H = Linear.y;
  I = Linear.z;
  J = Offset;
  numberOfQuadrics++;  // for parser statistics
}

// initializers
void Quadric::initSphere(const Vec3& Center, double Radius) {
  Vec3 linear;
  linear.scale(Center, -2.0);
  init(Vec3(1.0, 1.0, 1.0),  // Squared
       Vec3(0.0, 0.0, 0.0),  // Mixed
       linear,               // Linear
       Center.dot(Center) - Radius*Radius);
}

// along axis between p1 and p2
void Quadric::initCylinder(const Vec3& p1, const Vec3& p2, double Radius) {
  Vec3 a;   // cylinder axis

  Vec3 Squared, Mixed, Linear; // quadric parameters
  double Offset;

  a.x = p2.x - p1.x;
  a.y = p2.y - p1.y;
  a.z = p2.z - p1.z;
  
  Squared.x = a.y*a.y + a.z*a.z;
  Squared.y = a.x*a.x + a.z*a.z;
  Squared.z = a.x*a.x + a.y*a.y;

  Mixed.x = 2.0*(a.z*p1.y - a.z*p2.y);
  Mixed.y = 2.0*(a.x*p1.z - a.x*p2.z);
  Mixed.z = 2.0*(a.y*p1.x - a.y*p2.x);

  Linear.x = 2.0*((a.z*p1.z + a.y*p1.y)*p2.x - (a.z*p2.z + a.y*p2.y)*p1.x);
  Linear.y = 2.0*((a.z*p1.z + a.x*p1.x)*p2.y - (a.z*p2.z + a.x*p2.x)*p1.y);
  Linear.z = 2.0*((a.y*p1.y + a.x*p1.x)*p2.z - (a.y*p2.y + a.x*p2.x)*p1.z);
  
  Vec3 tmp = Vec3(p1.x*p2.z - p1.z*p2.x,
		  p2.y*p1.z - p2.z*p1.y,
		  p1.y*p2.x - p1.x*p2.y);

  Offset = tmp.dot(tmp) - (Radius*Radius*a.dot(a));

  init(Squared, Mixed, Linear, Offset);
}

void Quadric::initCone(const Vec3& p1, const Vec3& p2, double r1, double r2)
{
  // start from a 'squared distance to axis' cylindric quadric
  initCylinder(p1, p2, 0.0);

  // now fix up the coefficients for a cone

  double rd = (r1-r2);
  Vec3 a; // axis vector
  a.x = p2.x - p1.x;
  a.y = p2.y - p1.y;
  a.z = p2.z - p1.z;
  double den = 1.0/a.dot(a);
  
  Vec3 Squared, Mixed, Linear;
  double Offset;

  Squared.x = A - a.x*a.x*rd*rd*den;
  Squared.y = B - a.y*a.y*rd*rd*den;
  Squared.z = C - a.z*a.z*rd*rd*den;

  Mixed.x = D - 2.0*a.y*a.z*rd*rd*den;
  Mixed.y = E - 2.0*a.x*a.z*rd*rd*den;
  Mixed.z = F - 2.0*a.x*a.y*rd*rd*den;

  Linear.x = G + 2.0*((a.x*(r1*a.x + rd*p1.x) + ((a.y*p1.y + a.z*p1.z)*rd + r1*(a.z*a.z + a.y*a.y)))*rd*a.x)*den;
  Linear.y = H + 2.0*((a.y*(r1*a.y + rd*p1.y) + ((a.z*p1.z + a.x*p1.x)*rd + r1*(a.x*a.x + a.z*a.z)))*rd*a.y)*den;
  Linear.z = I + 2.0*((a.z*(r1*a.z + rd*p1.z) + ((a.x*p1.x + a.y*p1.y)*rd + r1*(a.y*a.y + a.x*a.x)))*rd*a.z)*den;

  double tmp = r1*a.dot(a) + rd*a.dot(p1);
  Offset = J - tmp*tmp*den;

  init(Squared, Mixed, Linear, Offset);
  numberOfQuadrics--;  // already incremented during cylinder init
}

// two parallel planes through p1 and p2, perpendicular to (p2 - p1)
void Quadric::initSlab(const Vec3& p1, const Vec3& p2)
{
  Vec3 a;  // axis vector
  a.x = p2.x - p1.x;
  a.y = p2.y - p1.y;
  a.z = p2.z - p1.z;

  Vec3 Squared, Mixed, Linear;
  double Offset;

  Squared.x = a.x*a.x;
  Squared.y = a.y*a.y;
  Squared.z = a.z*a.z;

  Mixed.x = 2*a.y*a.z;
  Mixed.y = 2*a.x*a.z;
  Mixed.z = 2*a.x*a.y;

  Linear.x = -a.x*(a.dot(p1) + a.dot(p2));
  Linear.y = -a.y*(a.dot(p1) + a.dot(p2));
  Linear.z = -a.z*(a.dot(p1) + a.dot(p2));
  
  Offset = a.dot(p1)*a.dot(p2);

  init(Squared, Mixed, Linear, Offset);
}

// algebraic product of two plane equations, i.e. XOR of two half-spaces
void Quadric::initPlanes(const Vec3& n1, double d1,
			 const Vec3& n2, double d2) {
  
  init(Vec3(n1.x*n2.x, n1.y*n2.y, n1.z*n2.z),
       Vec3(n1.y*n2.z + n1.z*n2.y,
	    n1.z*n2.x + n1.x*n2.z,
	    n1.y*n2.x + n1.x*n2.y),
       Vec3(d1*n2.x + d2*n1.x,
	    d1*n2.y + d2*n1.y,
	    d1*n2.z + d2*n1.z),
       d1*d2);

}

// put this quadric trough an affine linear transform
void Quadric::applyTransform(const Transform& trans) {
  double nA,nB,nC,nD,nE,nF,nG,nH,nI,nJ;        // new quadric

  //transform coefs
  const double& tA = trans.column[0].x;
  const double& tB = trans.column[1].x;
  const double& tC = trans.column[2].x;

  const double& tD = trans.column[0].y;
  const double& tE = trans.column[1].y;
  const double& tF = trans.column[2].y;

  const double& tG = trans.column[0].z;
  const double& tH = trans.column[1].z;
  const double& tI = trans.column[2].z;

  const double& tJ = trans.affine.x;
  const double& tK = trans.affine.y;
  const double& tL = trans.affine.z;

  nA = tA*(A*tA + E*tG) + tD*(B*tD + F*tA) + tG*(C*tG + D*tD);  // 14 FLOPs
  nB = tB*(A*tB + E*tH) + tE*(B*tE + F*tB) + tH*(C*tH + D*tE);
  nC = tC*(A*tC + E*tI) + tF*(B*tF + F*tC) + tI*(C*tI + D*tF);
  
  nD = (D*tF + 2.0*C*tI + E*tC)*tH + (F*tC + 2.0*B*tF + D*tI)*tE
    + (E*tI + F*tF + 2.0*A*tC)*tB;   // 23 FLOPs
  nE = (D*tF + 2.0*C*tI + E*tC)*tG + (F*tC + 2.0*B*tF + D*tI)*tD
    + (E*tI + F*tF + 2.0*A*tC)*tA;
  nF = (D*tE + 2.0*C*tH + E*tB)*tG + (D*tH + F*tB + 2.0*B*tE)*tD
    + (2.0*A*tB + E*tH + F*tE)*tA;
  
  nG = (2.0*C*tL + E*tJ + D*tK + I)*tG + (2.0*B*tK + H + F*tJ + D*tL)*tD
    + (E*tL + G + F*tK + 2.0*A*tJ)*tA;   // 26 FLOPs
  nH = (2.0*C*tL + E*tJ + D*tK + I)*tH + (2.0*B*tK + H + F*tJ + D*tL)*tE
    + (E*tL + G + F*tK + 2.0*A*tJ)*tB;
  nI = (2.0*C*tL + E*tJ + D*tK + I)*tI + (2.0*B*tK + H + F*tJ + D*tL)*tF
    + (E*tL + G + F*tK + 2.0*A*tJ)*tC;
  
  nJ = J + (I + C*tL)*tL + (D*tL + H + B*tK)*tK + (E*tL + G + F*tK + A*tJ)*tJ;
  // 18 FLOPs
  // total 3*(14 + 23 + 26) + 18 = 3*63 + 18 = 189 + 18 = 207 FLOPs

  init(Vec3(nA, nB, nC), Vec3(nD, nE, nF), Vec3(nG, nH, nI), nJ);
  numberOfQuadrics--;  // restore counter
}

// keep surface, but swap inside and outside
void Quadric::insideOut(void) {
  A = -A;
  B = -B;
  C = -C;
  D = -D;
  E = -E;
  F = -F;
  G = -G;
  H = -H;
  I = -I;
  J = -J;
}

// handle apex of a parabola
// internal use
static
inline void refineMinMax1D(double& minval, double& maxval,
			   double a, double b, double c,
			   double tmin, double tmax) {
  if(a != 0.0) {
    double pt = -b/(2.0*a);
    // clamp to limits
    pt = qmin(pt, tmax);
    pt = qmax(pt, tmin);
    
    double value = (a*pt + b)*pt + c;
    minval = qmin(minval, value);
    maxval = qmax(maxval, value);
  }
  return;
}

// handle apex of a 2D quadric and the two edges u = umin, u = umax and
// the one corner at u = umax, v = vmin
// internal use
static
void refineMinMax2D(double& minval, double& maxval,
		    double a, double b, double c, double d, double e, double f,
		    double umin, double umax, double vmin, double vmax) {

  // check 2D apex point
  double denominator = 4.0*a*b - c*c;
  double value;
  if (denominator != 0.0) {
    double pu = (c*e - 2.0*b*d)/denominator;
    double pv = (c*d - 2.0*a*e)/denominator;
    // clamp to limits
    pu = qmin(pu, umax);
    pu = qmax(pu, umin);
    pv = qmin(pv, vmax);
    pv = qmax(pv, vmin);

    value = (a*pu + c*pv + d)*pu + (b*pv + e)*pv + f;
    minval = qmin(minval, value);
    maxval = qmax(maxval, value);
  }

  // handle border at u = umin
  refineMinMax1D(minval, maxval,
                 b, c*umin + e, (a*umin + d)*umin + f,
		 vmin, vmax);
  // ... and border at u = umax
  // extract coefficients of parabola (for reuse)
  c = c*umax + e;
  a = (a*umax + d)*umax + f;
  refineMinMax1D(minval, maxval, b, c, a, vmin, vmax);
  
  // look at corner at u = umax, v = vmin
  value = (b*vmin + c)*vmin + a;
  minval = qmin(minval, value);
  maxval = qmax(maxval, value);
  
  return;
}

// intersect a quadric with an axis aligned box
// based on http://vectorizer.org/IntersectQuadricTetrahedron.html
//
// returns 2 (VoxInside) if the box is contained in the quadric,
//         1 (VoxOutside) if the box is completely outside the quadric
//         0 (VoxSurface) if the box overlaps the quadric surface
VoxClass Quadric::intersectVoxel(const GeObject*& Obj,
				 const Vec3& Min, const Vec3& Max) const {
  Obj = this;

  // handle the coordinate minimum corner and the coordinate maximum corner
  double value = evaluatePoint(Min);
  double minval = evaluatePoint(Max);
  double maxval = qmax(minval, value);
  minval = qmin(minval, value);

  // process the faces and edges of the box
  // (each of the six face calls handles two of the 12 edges and
  // one of the 8 corners; the remaining two corners are already done)
  
  // extract face at z = zmin as 2 dimensional quadric and process it
  // q(u,v) := Q(u,v,zmin)
  refineMinMax2D(minval, maxval,
    A, B, F, E*Min.z + G, D*Min.z + H, (C*Min.z + I)*Min.z + J,
    Min.x, Max.x, Min.y, Max.y);
  
  // q(u,v) := Q(u,v,zmax)
  refineMinMax2D(minval, maxval,
    A, B, F, E*Max.z + G, D*Max.z + H, (C*Max.z + I)*Max.z + J,
    Min.x, Max.x, Min.y, Max.y);
  
  // y = ymin (switched u and v to q(v,ymin,u) for symmetry)
  // q(u,v) := Q(v,ymin,u)
  refineMinMax2D(minval, maxval,
    C, A, E, D*Min.y + I, F*Min.y + G, (B*Min.y + H)*Min.y + J,
    Min.z, Max.z, Min.x, Max.x);
 
  // q(u,v) := Q(v,ymax,u)
  refineMinMax2D(minval, maxval,
    C, A, E, D*Max.y + I, F*Max.y + G, (B*Max.y + H)*Max.y + J,
    Min.z, Max.z, Min.x, Max.x);
 
  // x = xmin
  // q(u,v) := Q(xmin,u,v)
  refineMinMax2D(minval, maxval,
    B, C, D, F*Min.x + H, E*Min.x + I, (A*Min.x + G)*Min.x + J,
    Min.y, Max.y, Min.z, Max.z);
  
  // q(u,v) := Q(xmax,u,v)
  refineMinMax2D(minval, maxval,
    B, C, D, F*Max.x + H, E*Max.x + I, (A*Max.x + G)*Max.x + J,
    Min.y, Max.y, Min.z, Max.z);
  
  // find the 3D apex point
  double denominator = 8.0*A*B*C + 2.0*(D*E*F - A*D*D - B*E*E - C*F*F);
  // 16 FLOPs

  if (denominator != 0.0) {   // 1 FLOP
    double px = (2.0*B*E*I + D*D*G + 2.0*C*F*H - 4.0*B*C*G - D*F*I - D*E*H);
    double py = (E*E*H + 2.0*A*D*I + 2.0*C*F*G - 4.0*A*C*H - D*E*G - E*F*I);
    double pz = (2.0*B*E*G + 2.0*A*D*H + F*F*I - 4.0*A*B*I - E*F*H - D*F*G);
    // 3 * 20 = 60 FLOPs
    px /= denominator;  // three divides in total
    py /= denominator;
    pz /= denominator;
    
    // clamp position to limits
    px = qmin(px, Max.x);
    px = qmax(px, Min.x);
    py = qmin(py, Max.y);
    py = qmax(py, Min.y);
    pz = qmin(pz, Max.z);
    pz = qmax(pz, Min.z);

    value = evaluatePoint(Vec3(px, py, pz));
    minval = qmin(minval, value);
    maxval = qmax(maxval, value);
  }
  
  return VoxClass(((maxval < 0.0) << 1) + (minval > 0.0));  
}

// return closest hit beyond MinDist, or MAXNUM otherwise
double Quadric::intersectRay(Ray& Ray, double MinDist,
			     const GeObject*& Primitive) const {

  double values[5];  // storage space for results

  extractParabolaInternal(Ray, MinDist, values);
  
  Primitive = this;
  
  // beware, MinDist might have increased since cache was filled

  if (values[3] > MinDist) {  // if near hit is valid
    return values[3];         // return it
  }
  if (values[4] > MinDist) {  // if far hit is valid
    return values[4];         // return it
  }
  return MAXNUM;              // otherwise ray misses
}

// intersectRay which additionally returns coefficients of parabola
// Values[0] = square term
// Values[1] = linear term
// Values[2] = constant term
// Values[3] = near hit (or MAXNUM if invalid)
// Values[4] = far hit (or MAXNUM if invalid)
void Quadric::extractParabola(Ray& Ray, double MinDist,
			      double* Values) const {
  double values[5];   // storage space for results

  extractParabolaInternal(Ray, MinDist, values);
  
  // beware, MinDist might have increased since cache was filled
  if (values[3] <= MinDist) {
    // if near hit is invalid
    values[3] = values[4];  // replace with far hit
    values[4] = MAXNUM;
    
    if (values[3] <= MinDist) {
      // if far hit is invalid
      values[3] = MAXNUM;   // mark it as invalid
    }
  }
  
  // copy to destination buffer
  Values[0] = values[0];
  Values[1] = values[1];
  Values[2] = values[2];
  Values[3] = values[3];
  Values[4] = values[4];

  return;
}

__attribute__((always_inline))
void Quadric::extractParabolaInternal(Ray& Ray, double MinDist,
				      double* Values) const {
  double dx = Ray.dir.x;
  double dy = Ray.dir.y;
  double dz = Ray.dir.z;
  double px = Ray.base.x;
  double py = Ray.base.y;
  double pz = Ray.base.z;
  double TA, TB, TC;

  // global statistics: count quadric intersection tests
#if defined(STATISTICS)
  OSAtomicIncrement64(&numQuadric);  
#endif

  // find coefficients of one-dimensional polynomial along Ray
  // TA t^2 + TB t + TC

  // general case, with better accuracy
  TA = dx*(A*dx + F*dy) + dy*(B*dy + D*dz) + dz*(C*dz + E*dx);
  
  TB = (2.0*A*px + E*pz + F*py + G)*dx
    + (2.0*B*py + H + D*pz + F*px)*dy
    + (2.0*C*pz + I + E*px + D*py)*dz;

  TC = px*(A*px + E*pz + F*py + G) + py*(B*py + H + D*pz)
    + pz*(C*pz + I) + J;
  
  // pass back parabola
  Values[0] = TA;
  Values[1] = TB;
  Values[2] = TC;
  Values[3] = MAXNUM;   // initialize as two invalid hits
  Values[4] = MAXNUM;

  if (TA != 0.0) {  // check for degenerate case
    
    // divide out TA
    TA = 1.0/TA;
    TB *= TA;
    TC *= TA;
    
    // substitute TB for simpler formula
    TB *= -0.5;
    
    // now we have
    // x^2 - 2*TB*x + TC
    // solutions: TB +- sqrt(TB^2 - TC)

    double discriminant = TB*TB - TC;
    
    if (discriminant >= 0.0) {  // only hit when discriminant >= 0

      discriminant = sqrt(discriminant);

      // stadard formula
      // double near = TB - discriminant;
      // double far = TB + discriminant;
      //
      // numerical trickery
      // x1 = -(b+sign(b)*sqrt(D)) / 2*a 
      // x2 = c/(a*x1) 
      
      double near = TB + copysign(discriminant, TB);
      double far = TC/near;
      
      if (near <= MinDist) {
	near = MAXNUM;        // invalidate when in front of MinDist
      }
      if (far <= MinDist) {
	far = MAXNUM;
      }
      
      // sort near and far hits
      if (near > far) {
	discriminant = far;   // discriminant is just a temp here
	far = near;
	near = discriminant;
      }

      // pass back two (possibly invalidated) hits
      Values[3] = near;
      Values[4] = far;
    }
  } else {                // degenerate case: polynomial is linear
    if (TB != 0.0) {      // line is not parallel
      if (-TC/TB > MinDist) {
	// pass back one valid hit
	Values[3] = -TC/TB;
      }
    }
  }
  return;
}

// evaluate quadric equation at point in space
double Quadric::evaluatePoint(const Vec3& Point) const {
  double x = Point.x;
  double y = Point.y;
  double z = Point.z;
  
  return x*(A*x + F*y + G) + y*(B*y + D*z + H) + z*(C*z + E*x + I) + J;
}

// compute surface normal, i.e. unit length gradient at point in space
Vec3 Quadric::surfaceNormal(const Vec3& Point) const {
  double x = Point.x;
  double y = Point.y;
  double z = Point.z;
  Vec3 result;
  
  result.x = 2.0*A*x + F*y + E*z + G;
  result.y = 2.0*B*y + D*z + F*x + H;
  result.z = 2.0*C*z + E*x + D*y + I;

  result.normalize();

  return result; 
}
