/*************************************************************************
 *                        Boar 's Only A Raytracer                       *
 *                                                                       *
 *       Copyright (C) 2014  Holger Bettag  (hobold@vectorizer.org)      *
 *                                                                       *
 * This program is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation, either version 3 of the License, or     *
 * (at your option) any later version.                                   *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *************************************************************************/

#include <assert.h>
#include <stdio.h>

#if defined(STATISTICS)
#include <libkern/OSAtomic.h>
#endif

#include "trace.h"
#include "calc.h"
#include "simplelist.h"
#include "scene.h"

#if defined(STATISTICS)
extern int64_t numRays;  // statistics counters, defined in boar.c++
extern int64_t numShadow;
extern int64_t numBranch;
extern int64_t numLeaf;
#endif

// use this colour whenever a ray does not hit anything
const Vec3 BackgroundColour = Vec3(0.4, 0.6, 0.8);
//const Vec3 BackgroundColour = Vec3(0.5, 0.5, 0.5);
//const Vec3 BackgroundColour = Vec3(1.0, 1.0, 1.0);

// core routine, traces a ray and returns a colour
Vec3 TraceRay(Ray& ray, const Scene& scene,
#if defined(ANTIALIAS)
	      uintptr_t& Hash,
#endif  // defined(ANTIALIAS)
	      int Depth) {
  Vec3 result = BackgroundColour;
  GeObject* object;
  double rayT;       // ray parameter of hit
  Vec3 normal;       // surface normal of hit

  int splits = 0;

  // global statistics: count view rays
#if defined(STATISTICS)
  OSAtomicIncrement64(&numRays);
#endif
  
  object = ClosestHit(ray, scene, 0.0, rayT, splits);
  
  if (object != NULL) {  // if we did hit something, light it
    Material* mat = object->getMaterial();
    //__builtin_prefetch(mat);  // not beneficial
    Vec3 pos;            // position of hit
    
    pos.scale(ray.dir, rayT);
    pos.add(ray.base);
    
    /*
    pos.x = ray.base.x + rayT*ray.dir.x;
    pos.y = ray.base.y + rayT*ray.dir.y;
    pos.z = ray.base.z + rayT*ray.dir.z;
    */

    normal = object->surfaceNormal(pos);

    
    if (normal.len() == 0.0) {
      //fprintf(stderr,"%20.18f\n", normal.len());
      normal.x = 1.0;
    }
    

    if(normal.dot(ray.dir) > 0.0) {  // does the normal face the ray?
      normal.scale(-1.0);
      //return Vec3(1.0, 1.0, 1.0);  // for debugging
    }

    assert(normal.len() > 0.9999);
    assert(normal.len() < 1.0001);
    

#if defined(ANTIALIAS)
    // mix lighted surfaces to light path hash
    Hash ^= uintptr_t(object);
    Hash = (Hash << 17) | (Hash >> ((sizeof(uintptr_t) << 3) - 17));
#endif  // defined(ANTIALIAS)
    
    result = mat->lighting(ray.dir, pos, normal, scene,
#if defined(ANTIALIAS)
			   Hash,
#endif  // defined(ANTIALIAS)
			   Depth);
  }

#if defined(VOXELS)  
  // for debugging: color shift based on voxel traversal
  if (splits > 20) {
    splits = 20;      // clamp
  }
  result.scale(double(20 - splits)/20.0);
  result.y += double(splits)/20.0;
#endif

  return result;
}

// INTERNAL VERSION WITHOUT SPATIAL SUBDIVISION
// utility routine, traces a ray and returns closest hit beyond minDist
// corresponding ray parameter is returned via reference
// in case of miss, GeObject is NULL, and ray parameter is MAXNUM
GeObject* rawClosestHit(Ray& ray, const List* objects,
			double minDist, double& where) {

  const GeObject* result = NULL;
  const GeObject* curObject;
  
  double dist = MAXNUM;
  double curDist;

  const ListNode* objList;
  if (objects != NULL) {
    objList = objects->firstNode();
    //__builtin_prefetch(objList);  // not beneficial here
  } else {
    objList = NULL;
  }

  while(objList != NULL) {

    curObject = (GeObject*)objList->getData();
    //__builtin_prefetch(curObject);  // not beneficial here
    curDist = curObject->intersectRay(ray, minDist, curObject);
    if (curDist < dist) {
      result = curObject;
      dist = curDist;
    }
    objList = objList->nextNode();
  }
  
  where = dist;
  return (GeObject*)result;
}

/*
// uses spatial subdivision
// utility routine, traces a ray and returns closest hit beyond minDist
// corresponding ray parameter is returned via reference
// in case of miss, GeObject is NULL, and ray parameter is MAXNUM
GeObject* ClosestHitOld(const Ray& ray, const Scene& scene,
			double minDist, double& where, int& Splits) {
  Voxel* stackVoxel[SSDMAXDEPTH];   // stack of voxels to be traversed
  double stackDist[SSDMAXDEPTH];    // corresponding minDist values
  int stackptr = 0;                 // stack pointer

  Voxel* curVoxel = scene.root;  // start traversal at root voxel
  double dist;
  GeObject* object;

  //Splits = 0;  // value in case of total miss

  dist = curVoxel->intersectRay(ray, minDist);
  
  if (MAXNUM == dist) {  // root voxel missed => ray misses
    where = MAXNUM;
    return NULL;
  }
  
  minDist = qmax(dist, minDist);  // advance minDist if needed

  do {
    // precondition: ray intersects curVoxel
    if (curVoxel == NULL) {
      fprintf(stderr,"trace(): curVoxel is NULL (0).\n");
      exit(0);
    }
    
    if (curVoxel->isLeaf) {
      // current voxel is a leaf, so process the surfaces within
      object = rawClosestHit(ray, curVoxel->getSurfaces(), minDist, where);
      // only accept hits within current voxel
      Vec3 loc = ray.evaluate(where);  // location of hit
      if ((loc.x >= curVoxel->min.x) && (loc.x <= curVoxel->max.x)) {
	if ((loc.y >= curVoxel->min.y) && (loc.y <= curVoxel->max.y)) {
	  if ((loc.z >= curVoxel->min.z) && (loc.z <= curVoxel->max.z)) {
	    //Splits = long(curVoxel) % 65537;
	    return object;  // hit is inside current voxel, so it is valid
	  }
	}
      }
      // if hit is invalid, pop from stack
      if (stackptr > 0) {
	curVoxel = stackVoxel[--stackptr];
	if (curVoxel == NULL) {
	  fprintf(stderr,"trace(): curVoxel is NULL (1).\n");
	}
	minDist = stackDist[stackptr];
	Splits++;
	continue;  // goto top
      } else {
	break;  // empty stack means we are done, ray misses
      }
    } else {
      // current voxel is a branch, intersect children
      Voxel* leftVoxel = curVoxel->getLeftChild();
      Voxel* rightVoxel = curVoxel->getRightChild();
      
      // pathological case: two empty children
      // should now be fixed in tree construction
      if((NULL == leftVoxel) && (NULL == rightVoxel)) {
	fprintf(stderr,"trace(): two NULL children.\n");
	exit(0);
      }

      // test all existing children
      double lDist = MAXNUM, rDist = MAXNUM;
      if (leftVoxel != NULL) {
	lDist = leftVoxel->intersectRay(ray, minDist);
      }
      if (rightVoxel != NULL) {
	rDist = rightVoxel->intersectRay(ray, minDist);
      }
      
      // sort children as near, far
      // (miss is MAXNUM, so near is always hit, and a miss is always far)
      if (lDist < rDist) {
	curVoxel = leftVoxel;            // proceed into near child
	if (curVoxel == NULL) {
	  fprintf(stderr,"trace(): curVoxel is NULL (3).\n");
	}
	minDist = qmax(lDist, minDist);  // advance minDist
	if (rDist < MAXNUM) {            // push far child on stack if valid
	  if (NULL == rightVoxel) {
	    fprintf(stderr,"trace(): pushed NULL far (right) child.\n");
	  }
	  stackVoxel[stackptr] = rightVoxel;
	  stackDist[stackptr++] = qmax(rDist, minDist);
	}
      } else {
	if (rDist < MAXNUM) {
	  curVoxel = rightVoxel;           // proceed into near child
	  if (curVoxel == NULL) {
	    fprintf(stderr,"trace(): curVoxel is NULL (4).\n");
	  }
	  minDist = qmax(rDist, minDist);  // advance minDist
	  if (lDist < MAXNUM) {            // push far child on stack if valid
	    if (NULL == rightVoxel) {
	      fprintf(stderr,"trace(): pushed NULL far (left) child.\n");
	    }
	    stackVoxel[stackptr] = leftVoxel;
	    stackDist[stackptr++] = qmax(lDist, minDist);
	  }
	} else {
	  // the non-empty child is missed, pop from stack
	  if (stackptr > 0) {
	    curVoxel = stackVoxel[--stackptr];
	    if (curVoxel == NULL) {
	      fprintf(stderr,"trace(): curVoxel is NULL (5).\n");
	    }
	    Splits++;
	    minDist = stackDist[stackptr];
	    continue;  // goto top
	  } else {
	    break;  // empty stack means we are done, ray misses
	  }
	}
      }
    }  
  } while (true);
  
  where = MAXNUM;
  return NULL;
}
*/



// uses spatial subdivision, new KdTree structure
// utility routine, traces a ray and returns closest hit beyond minDist
// corresponding ray parameter is returned via reference
// in case of miss, GeObject is NULL, and ray parameter is MAXNUM
GeObject* ClosestHit(Ray& ray, const Scene& scene,
		     double minDist, double& where, int& Splits) {

  double tnear = -MAXNUM, tfar = MAXNUM;  // current valid ray interval
  GeObject* object = NULL;                // result variable
  where = MAXNUM;
  
  Vec3 reciDir;    // reciprocal ray direction
  for (int i = 2; i >= 0; i--) {
    if (ray.dir[i] != 0.0) {
      reciDir[i] = 1.0/ray.dir[i];
    } else {
      reciDir[i] = 0.0;
    }
  }

    
  // for debugging
  if (0.0 == ray.dir.len()) {
    fprintf(stderr, "ClosestHit(): ray.dir is 0\n");
  }
  

  // tnear and tfar have been determined
  
      
  // statistics
#if defined(STATISTICS)
  OSAtomicIncrement64(&numLeaf);
#endif
      
  object = rawClosestHit(ray, &scene.objects, minDist, where);

  if (where <= tfar) {  // if hit is inside current node ...
    assert(where >= tnear);
    return object;      // ... then it is the valid result
  }
  
  where = MAXNUM;   // if we get here, ray misses
  return NULL;
}

// utility routine, traces a ray and returns any hit (for shadow calculations)
// in case of miss, GeObject is NULL
GeObject* AnyHit(Ray& ray, const Scene& scene, double maxDist) {

  // unoptimized variant relying on ClosestHit()

  GeObject* result;
  double where;
  int splits = 0;
  
  // global statistics: count shadow rays
#if defined(STATISTICS)
  OSAtomicIncrement64(&numShadow);
#endif

  result = ClosestHit(ray, scene, 0.0, where, splits);
  
  if (where > maxDist)
    result = NULL;
  
  return result;
}
