/*************************************************************************
 *                        Boar 's Only A Raytracer                       *
 *                                                                       *
 *       Copyright (C) 2014  Holger Bettag  (hobold@vectorizer.org)      *
 *                                                                       *
 * This program is free software: you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation, either version 3 of the License, or     *
 * (at your option) any later version.                                   *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *************************************************************************/

#include <math.h>

#include "transform.h"


Transform::Transform(void) {
  column[0] = Vec3(1.0, 0.0, 0.0);
  column[1] = Vec3(0.0, 1.0, 0.0);
  column[2] = Vec3(0.0, 0.0, 1.0);
  affine = Vec3(0.0, 0.0, 0.0);
}


Transform::Transform(const Vec3& Col1, const Vec3& Col2, const Vec3& Col3,
		     const Vec3& Affine) {
  column[0] = Col1;
  column[1] = Col2;
  column[2] = Col3;
  affine = Affine;
}


void Transform::RotateX(double angle) {
  double sv, cv;
  angle = -M_PI*angle/180.0;  // invert & convert deg2rad

  sv = sin(angle);
  cv = cos(angle);

  Transform temp(Vec3(1.0, 0.0, 0.0),
		 Vec3(0.0,  cv,  sv),
		 Vec3(0.0, -sv,  cv),
		 Vec3(0.0, 0.0, 0.0));

  Apply(temp);
}


void Transform::RotateY(double angle) {
  double sv, cv;
  angle = -M_PI*angle/180.0;  // invert & convert deg2rad

  sv = sin(angle);
  cv = cos(angle);

  Transform temp(Vec3( cv, 0.0, -sv),
		 Vec3(0.0, 1.0, 0.0),
		 Vec3( sv, 0.0,  cv),
		 Vec3(0.0, 0.0, 0.0));

  Apply(temp);
}


void Transform::RotateZ(double angle) {
  double sv, cv;
  angle = -M_PI*angle/180.0;  // invert & convert deg2rad

  sv = sin(angle);
  cv = cos(angle);

  Transform temp(Vec3( cv,  sv, 0.0),
		 Vec3(-sv,  cv, 0.0),
		 Vec3(0.0, 0.0, 1.0),
		 Vec3(0.0, 0.0, 0.0));

  Apply(temp);
}


void Transform::Scale(const Vec3& factors) {
  Transform temp(Vec3(1.0/factors.x, 0.0, 0.0),  // invert
		 Vec3(0.0, 1.0/factors.y, 0.0),
		 Vec3(0.0, 0.0, 1.0/factors.z),
		 Vec3(0.0, 0.0, 0.0));

  Apply(temp);
}


void Transform::Translate(const Vec3& offset) {
  Transform temp;

  temp.affine.x = -offset.x;  // invert
  temp.affine.y = -offset.y;
  temp.affine.z = -offset.z;

  Apply(temp);  // inefficient, but not in a critical path
}


void Transform::Apply(const Transform& t) {
  Vec3 col0, col1, col2, aff;  // temporaries
  
  col0 = MatxVec(t.column[0]);
  col1 = MatxVec(t.column[1]);
  col2 = MatxVec(t.column[2]);

  aff = ApplyTo(t.affine);

  column[0] = col0;
  column[1] = col1;
  column[2] = col2;

  affine = aff;
}


Vec3 Transform::ApplyTo(const Vec3& vec) const {
  Vec3 result;

  result = MatxVec(vec);
  result.add(affine);

  return result;
}


Vec3 Transform::MatxVec(const Vec3& vec) const {
  Vec3 temp, result;
  
  temp = Vec3(column[0].x, column[1].x, column[2].x);
  result.x = vec.dot(temp);
  
  temp = Vec3(column[0].y, column[1].y, column[2].y);
  result.y = vec.dot(temp);

  temp = Vec3(column[0].z, column[1].z, column[2].z);
  result.z = vec.dot(temp);

  return result;
}
